
##-- imports
from __future__ import annotations

import logging as logmod
from collections import defaultdict
from importlib.resources import files
from string import Template, ascii_uppercase

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##-- data
tex_path        = files(TEX_loc)
HEADER_PAT      = Template((tex_path / "header_pattern").read_text())
GANTT_BAR       = Template((tex_path / "gantt_bar.tex").read_text())
GANTT_CHART     = Template((tex_path / "gantt_chart.tex").read_text())
GANTT_MILESTONE = Template((tex_path / "gantt_milestone.tex").read_text())
GANTT_PRELUDE   = Template((tex_path / "gantt_prelude.tex").read_text())

TERM            = Template((tex_path / "term.tex").read_text())
TERM_BODY       = Template((tex_path / "term_body.tex").read_text())
##-- end data

from jgdv.files.tex.base import TexBuilder_i

class GanttTexBuilder(TexBuilder_i):
    """
    InstalGanttTracer
    Implementation of Reporter for gantt output.

    Milestones are events
    Bars are fluents
    """

    def header(self):
        pass

    def footer(self):
        pass

    def prepare2(self):
        for t in range(1, len(observed) + 1):
            for x in occurred[t][:]:
                l = (str(x.arguments[0]) + ": " +
                        str(x.arguments[1])).replace('_', '\_')
                print("\\ganttmilestone{{{l}}}{{{f}}}\\ganttnewline"
                        .format(l=l, f=t - 1), file=tfile)

        facts = invert(holdsat)
        keys = sorted(facts, key=lambda atom: atom.arguments[0].name)
        for f in keys:
            print(
                r"\begin{ganttchart}[hgrid,vgrid,canvas/.style={draw=none},bar/.append style={fill=gray},x unit=0.5cm,y unit chart=0.5cm]{0}" +
                "{{{t}}}\n".format(t=len(observed) + 1), file=tfile)
            i = facts[f][0]
            l = (str(f.arguments[0]) + ": " +
                    str(f.arguments[1])).replace('_', '\_')
            print("\\ganttbar{{{label}}}{{{start}}}{{{finish}}}"
                    .format(label=l, start=i, finish=i), file=tfile)

    def render_term(self, term, inst=None) -> str:
        """
        Render a passed in ASTTerm into a useable latex string.
        Adapts to handle institutional terms and non.
        """
        body = ""
        if bool(term.params):
            body = self.expand(TERM_BODY, body=",".join(str(x) for x in term.params))

        match inst:
            case None:
                return self.expand(TERM,
                                   head=term.value,
                                   body=body.strip()).strip().replace("_", "\_").strip()
            case _:
                return self.expand(INST_TERM,
                                   head=term.value,
                                   body=body.strip(),
                                   inst=inst).strip().replace("_", "\_").strip()

    def render_fluents(self, state, prior_state=None) -> tuple[set, set, set]:
        """
        Separate out fluents in a state into
        {holding}, {initiated}, and {terminated} sets.

        """
        init   = {self.render_term(x.params[0]) for x in state.fluents if prior_state is None or x not in prior_state}
        holds  = {self.render_term(x.params[0]) for x in prior_state.fluents if x in state} if prior_state is not None else {}
        termin = {self.render_term(x.params[0]) for x in prior_state.fluents if x not in state} if prior_state is not None else {}

        init_render   = {self.expand(FLUENT, mod=BOLD,  term=term).strip() for term in init}
        holds_render  = {self.expand(FLUENT, mod="",    term=term).strip() for term in holds}
        termin_render = {self.expand(FLUENT, mod=SOUT, term=term).strip() for term in termin}

        return sorted(init_render), sorted(holds_render), sorted(termin_render)

    def render_milestones(self, trace, title="Observed Events", caption="Tracking the events in the trace") -> str:
        """
        Convert the trace to a gantt chart of milestones,
        where each milestone is an event observed in the trace
        """
        content = []

        for state in trace:
            for event in state.observed:
                event_s : str = self.render_term(event.params[0])
                milestone = self.expand(GANTT_MILESTONE,
                                        text=event_s,
                                        num=state.timestep,
                                        )
                content.append(milestone)

        chart = self.expand(GANTT_CHART,
                            content="\n".join(sorted(content)),
                            num=len(trace),
                            numPlusOne=len(trace)+1,
                            title=title,
                            caption=caption
                            )

        return chart

    def render_bars(self, trace, title="Fluents", caption="Gantt Chart for Fluent lives") -> str:
        """
        Convert the trace to a gantt chart,
        where the durations of each fluent are displayed
        """
        content = []
        intervals = trace.fluent_intervals()
        ## a a bar for the fact from start to finish
        for term, start, fin in intervals:
            term_s : str = self.render_term(term.params[0])
            content.append(self.expand(GANTT_BAR,
                                       text=term_s,
                                       start=str(start),
                                       end=str(fin)))

        # wrap in a chart, add to results
        chart = self.expand(GANTT_CHART,
                            content="\n".join(sorted(content)),
                            num=len(trace),
                            numPlusOne=len(trace)+1,
                            title=title,
                            caption=caption,
                            )
        return chart

    def render_bars_by_institution(self, trace) -> list[str]:
        """
        Convert the trace to n gantt charts,
        one for each institution
        """
        institution_list = trace.metadata['institutions']
        charts           = []

        for inst in institution_list:
            rejections = [x for x in institution_list if x != inst]
            f_trace    = trace.filter(allow=[inst], reject=rejections)
            inst_chart = self.render_bars(f_trace,
                                          title=f"{inst} Fluents",
                                          caption=f"Fluent Lives in institution {inst}")
            charts.append(inst_chart)

        return charts

    def trace_to_file(self, trace, path):
        self.clear()
        all_gantts = []
        # build event gantt
        all_gantts.append(self.render_milestones(trace))
        # build fluent gantt
        all_gantts.append(self.render_bars(trace))
        all_gantts += self.render_bars_by_institution(trace)

        # Insert prelude
        self.insert(GANTT_PRELUDE,
                    body="\n".join(all_gantts))

        with open(path, 'w') as f:
            f.write("\n".join(self._compiled_text))



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from functools import partial
from math import atan2, copysign

import cairo_utils as cu
import numpy as np
from cairo_utils.constants import (BEZIER_DATA_LEN, CIRCLE_DATA_LEN, DELTA,
                                   EPSILON, HALFDELTA, LINE_DATA_LEN,
                                   NODE_RECIPROCAL, PI, QUARTERPI,
                                   SAMPLE_DATA_LEN, TOLERANCE, TWOPI,
                                   IntersectEnum)
from numpy import cos, pi, sin
from scipy.interpolate import splev, splprep

def sample_circle(xyrs, n, sort_rads=True, sort_radi=True, easing=None, random=None):
    """
    https://stackoverflow.com/questions/5837572
    """
    if random is None:
        random = np.random.random

    xyrs_r = xyrs.reshape((-1, 1, CIRCLE_DATA_LEN)).repeat(n, axis=1)

    r = scale_ndarray(np.sqrt(random((n, 1))), xyrs_r[0,:,4:])
    if sort_radi:
        r.sort(axis=1)

    theta = scale_ndarray(random((n, 1)), xyrs_r[0,:,2:4])
    if sort_rads:
        theta.sort(axis=1)

    r_shaped = r.reshape((xyrs_r.shape[0],n,1))
    theta_shaped = theta.reshape((xyrs_r.shape[0],n,1))

    rot = np.dstack((np.cos(theta_shaped), np.sin(theta_shaped)))
    r_x_rot = r_shaped * rot

    result = xyrs_r[:,:,:2] + r_x_rot
    flattened = result.reshape((-1, 2))
    return flattened

def get_circle_3p(p1, p2, p3, arb_intersect=20000):
    """
    Given 3 points, treat them as defining two chords on a circle,
    intersect them to find the centre, then calculate the radius
    Thus: circumcircle
    """
    #pylint: disable=too-many-locals
    assert(all([isinstance(x, np.ndarray) for x in [p1, p2, p3]]))
    sorted_points = sort_coords(np.array([p1, p2, p3]))
    p1 = sorted_points[0]
    p2 = sorted_points[1]
    p3 = sorted_points[2]

    arb_height = arb_intersect
    #mid points and norms:
    m1 = get_midpoint(p1, p2)
    n1 = get_bisector(m1, p2)
    m2 = get_midpoint(p2, p3)
    n2 = get_bisector(m2, p3)
    #extended norms:
    v1 = m1 + (1 * arb_height * n1)
    v2 = m2 + (1 * arb_height * n2)
    v1_i = m1 + (-1 * arb_height * n1)
    v2_i = m2 + (-1 * arb_height * n2)
    #resulting lines:
    l1 = np.row_stack((m1, v1))
    l2 = np.row_stack((m2, v2))
    l1_i = np.row_stack((m1, v1_i))
    l2_i = np.row_stack((m2, v2_i))
    #intersect extended norms:
    #in the four combinations of directions
    i_1 = intersect(l1, l2)
    i_2 = intersect(l1_i, l2_i)
    i_3 = intersect(l1, l2_i)
    i_4 = intersect(l1_i, l2)
    #get the intersection:
    the_intersect = [x for x in [i_1, i_2, i_3, i_4] if x is not None]
    if the_intersect is None or not bool(the_intersect):
        return None
    r1 = get_distance(p1, the_intersect[0])
    r2 = get_distance(p2, the_intersect[0])
    r3 = get_distance(p3, the_intersect[0])

    #a circle only if they are have the same radius
    if np.isclose(r1, r2) and np.isclose(r2, r3):
        return (the_intersect[0], r1)
    else:
        return None

def get_lowest_point_on_circle(centre, radius):
    """
    given the centre of a circle and a radius, get the lowest y point on that circle
    """
    return centre - np.array([0, radius])

def in_circle(centre, radius, points):
    """ Test a set of points to see if they are within a circle's radius """
    d = get_distance_raw(centre, points)
    return d < pow(radius, 2)



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
# import more_itertools as mitz
# from boltons import
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@dataclass
class TomlLocation:
    """ A representation of a location specified in toml.
      This is the single point of truth for converting a location specified in toml
      to preserve its metadata.
      eg: if its a file, protected, etc

    in toml, will be:
      {file="..."},
      {protected="..."}

    or for combinations:
      {loc="...", protected=true...}
    """
    key  : str          = field()
    base : pl.Path      = field()
    meta : LocationMeta = field(default=LocationMeta.default)

    @staticmethod
    def build(key:str, data:dict|str, base:pl.Path=None):
        result = None
        match data:
            case str():
                result = TomlLocation(key=key, base=(base or pl.Path(data)))
            case pl.Path():
                result = TomlLocation(key=key, base=(base or data))
            case TomlLocation():
                result = TomlLocation(key=data.key, base=(base or data.base), meta=data.meta)
            case dict() if base is not None:
                meta   = LocationMeta.build({x:y for x,y in data.items() if x != "loc"})
                result = TomlLocation(key=key, base=base, meta=meta)
            case dict() if 'loc'in data:
                meta   = LocationMeta.build({x:y for x,y in data.items() if x != "loc"})
                result = TomlLocation(key=key, base=pl.Path(data['loc']), meta=meta)

        return result

    def check(self, meta:LocationMeta) -> bool:
        return meta in self.meta



"""
Trie Class for bookmarks
"""
##-- imports
from __future__ import annotations

import logging as logmod
from dataclasses import InitVar, dataclass, field
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)
##-- end imports

from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

logging = logmod.getLogger(__name__)

@dataclass
class Trie:
    """ Main Trie Access class """
    data : InitVar[List[Any]] = field(default=None)

    root             : Dict[Any, Any] = field(init=False, default_factory=dict)
    leaves           : List[Any]      = field(init=False, default_factory=list)
    query_keys       : Dict[Any, Any] = field(init=False, default_factory=dict)
    query_key_counts : Dict[Any, int] = field(init=False, default_factory=dict)

    __repr__ = __str__

    def __post_init__(self, data=None):
        if data is not None:
            for x in data:
                self.insert(x)

    def __len__(self):
        return len(self.leaves)

    def __str__(self):
        return "Trie: {}, {}".format(len(self), len(self.query_keys))

    def get_tuple_list(self):
        results = []
        for x in self.leaves:
            results += x.get_tuple_list()

        return results

    def insert(self, data):
        """ Insert a bookmark into the trie,
        based on url components """
        assert(isinstance(data, Bookmark))

        #Get components of the url
        p_url = urlparse(data.url)
        trie_path = [p_url.scheme, p_url.netloc] + p_url.path.split('/')
        f_trie_path = [x for x in trie_path if x]

        query = parse_qs(p_url.query)

        #find the leaf
        current_child = self.root
        for x in f_trie_path:
            if x not in current_child:
                current_child[x] = {}
            current_child = current_child[x]

        #insert into the leaf, merging tag sets
        if '__leaf' not in current_child:
            new_leaf = Leaf()
            current_child['__leaf'] = new_leaf
            self.leaves.append(new_leaf)

        leaf = current_child['__leaf']
        leaf_node = leaf.insert(data.name, p_url, data.tags, query, data.url)

        for k in query.keys():
            if k not in self.query_keys:
                self.query_keys[k] = (data.url, leaf_node.reconstruct(k))
                self.query_key_counts[k] = 0
            self.query_key_counts[k] += 1

    def filter_queries(self, query_set):
        for x in self.leaves:
            x.filter_queries(query_set)

    def org_format_queries(self):
        """
        Output a list of org links, with original URLs,
        and URL's minus a query parameter.
        Used to find out which parameters can be filtered from links
        """
        result = []
        for key, url_pair in self.query_keys.items():
            count = self.query_key_counts[key]
            result.append("** ({}) {}\n  [[{}][original]]\n  [[{}][filtered]]".format(count,
                                                                                      key,
                                                                                      url_pair[0],
                                                                                      url_pair[1]))
        return "\n".join(result)



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import tracemalloc
import datetime

Seconds  : TypeAlias = int
DateTime : TypeAlias = datetime.datetime

class HumanNum:
    """
    Simple human number related functions
    """

    @staticmethod
    def human_sizes(val, format=False):
        return tracemalloc._format_size(val, format)

    @staticmethod
    def round_time(dt:DateTime=None, roundTo:Seconds=60) -> DateTime:
        """Round a datetime object to any time lapse in seconds
        dt : datetime.datetime object, default now.
        roundTo : Closest number of seconds to round to, default 1 minute.
        Author: Thierry Husson 2012 - Use it as you want but don't blame me.
        from: https://stackoverflow.com/questions/3463930
        """
        dt       = dt or datetime.datetime.now()
        seconds  = (dt.replace(tzinfo=None) - dt.min).seconds
        rounding = (seconds+roundTo/2) // roundTo * roundTo
        return dt + datetime.timedelta(0,rounding-seconds,-dt.microsecond)



#!/usr/bin/env python3
##-- imports
from __future__ import annotations

import abc
import logging as logmod
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from os.path import splitext
from re import Pattern
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

logging = logmod.getLogger(__name__)
SC      = SimpleLogColour

def print_header(header):
    print("\n--------------------")
    print(f"{SC.green(header)}: ")
    print("--------------------")

def print_class_colour(heading, cls):
    print(f"{heading}: {SC.yellow(cls.__module__)}.{SC.green(cls.__class__.__name__)}")
    if bool(cls.__doc__):
        print(SC.blue(cls.__doc__))

    print()
    bases = "\n\t".join(f"{SC.yellow(x.__module__)}.{SC.green(x.__qualname__)}" for x in cls.__class__.__bases__)
    print(f"Instance of:\n\t{bases}")

def print_module_colour(mod):
    result = []
    if bool(len(mod.dsl_fragments)):
        result.append(SC.green(f"{len(mod.dsl_fragments)} DSL"))
    else:
        result.append(SC.red(f"{len(mod.dsl_fragments)} DSL"))
    if bool(len(mod.semantics)):
        result.append(SC.green(f"{len(mod.semantics)} Sem"))
    else:
        result.append(SC.red(f"{len(mod.semantics)} Sem"))
    if bool(len(mod.operators)):
        result.append(SC.green(f"{len(mod.operators)} Op"))
    else:
        result.append(SC.red(f"{len(mod.operators)} Op"))
    if bool(len(mod.printers)):
        result.append(SC.green(f"{len(mod.operators)} Pr"))
    else:
        result.append(SC.red(f"{len(mod.operators)} Pr"))

    source = splitext(mod.source)
    print(f"\t({' | '.join(result)} : {source[0]}{SC.blue(source[1])})")

def print_handler_system(system):
    print("\nComponents: (no. of handlers : signal : flags)")
    sorted_specs = sorted([(len(y), y, x) for x,y in system.handler_specs.items()], reverse=True, key=lambda x: x[0])
    for spec_len, spec, signal in sorted_specs:
        spec_len = SC.green(f"{spec_len:<4}") if bool(spec_len) else SC.red(f"{spec_len:<4}")
        spec_flags = [x.name for x in spec.flags]
        flag_str = f" : {', '.join(spec_flags)}" if bool(spec_flags) else ""
        print(f"\t{spec_len} : {SC.blue(signal):<40}{flag_str}")

    print("\nHandlers not attached to a Signal: ")
    for signal in sorted((x.signal for x in system.loose_handlers)):
        print(f"\t{SC.blue(signal)}")

def two_part_split(the_string):
    pair = splitext(the_string)
    return f"{pair[0]}{SC.blue(pair[1])}"



"""


"""
##-- imports
from __future__ import annotations

import abc
import collections.abc
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
##-- end imports



#!/usr/bin/env python4
"""

"""
##-- imports
from __future__ import annotations

import logging as logmod
import warnings
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast)
##-- end imports
logging = logmod.root

import pytest

@pytest.mark.parametrize("ctor", [DootFlexibleParser])
class TestArgParser:

    def test_initial(self, ctor):
        parser = ctor()
        assert(isinstance(parser, ArgParser_i))

    def test_cmd(self, ctor, mocker):
        cmd_mock                   = mock_parse_cmd(name="list")
        parser                     = ctor()

        result                     = parser.parse(["doot", "list"],
            doot_specs=[], cmds={"list": cmd_mock}, tasks={}
            )

        assert(result.on_fail(False).head.name() == "doot")
        assert(result.on_fail(False).cmd.name()  == "list")

    def test_cmd_args(self, ctor, mocker):
        cmd_mock  = mock_parse_cmd(params=[DootParamSpec(name="all")])
        parser    = ctor()
        result    = parser.parse([
            "doot", "list", "-all"
                               ],
            doot_specs=[], cmds={"list": cmd_mock}, tasks={}
            )
        assert(result.on_fail(False).head.name() == "doot")
        assert(result.on_fail(False).cmd.name() == "list")
        assert(result.on_fail(False).cmd.args.all() == True)

    def test_cmd_arg_fail(self, ctor, mocker):
        cmd_mock                   = mock_parse_cmd(params=[DootParamSpec(name="all")])

        parser = ctor()
        with pytest.raises(Exception):
            parser.parse([
                "doot", "list", "-all", "-bloo"
            ],
            doot_specs=[], cmds={"list": cmd_mock}, tasks={}
        )

    def test_cmd_then_task(self, ctor, mocker):
        cmd_mock                   = mock_parse_cmd(params=[DootParamSpec(name="all")])
        task_mock                  = mock_parse_task(params=[{"name":"all"}])

        parser = ctor()
        result = parser.parse([
            "doot", "list", "blah"
                               ],
            doot_specs=[], cmds={"list": cmd_mock}, tasks={"blah": task_mock},
            )
        assert(result.on_fail(False).head.name() == "doot")
        assert(result.on_fail(False).cmd.name() == "list")
        assert("blah" in result.tasks)

    def test_cmd_then_complex_task(self, ctor, mocker):
        cmd_mock  = mock_parse_cmd(params=[DootParamSpec(name="all")])
        task_mock = mock_parse_task(params=[{"name":"all"}])

        parser    = ctor()
        result    = parser.parse([
            "doot", "list", "blah::bloo.blee"
                               ],
            doot_specs=[], cmds={"list": cmd_mock}, tasks={"blah::bloo.blee": task_mock},
            )
        assert(result.on_fail(False).head.name() == "doot")
        assert(result.on_fail(False).cmd.name() == "list")
        assert( "blah::bloo.blee" in result.tasks)

    def test_task_args(self, ctor, mocker):
        """ check tasks can recieve args """
        cmd_mock                                    = mock_parse_cmd()
        task_mock                                   = mock_parse_task(params=[{"name":"all"}])

        parser                                      = ctor()
        result                                      = parser.parse([
            "doot", "basic::list", "-all"
                               ],
            doot_specs=[], cmds={"run": cmd_mock}, tasks={"basic::list": task_mock},
            )
        assert(result.on_fail(False).head.name() == "doot")
        assert(result.on_fail(False).cmd.name() == "run")
        assert(bool(result.on_fail(False).tasks["basic::list"]()))
        assert(result.on_fail(False).tasks["basic::list"].all() == True)

    def test_task_with_name_spaces(self, ctor, mocker):
        cmd_mock                   = mock_parse_cmd()
        task_mock                  = mock_parse_task(params=[{"name":"all"}])

        parser = ctor()
        result = parser.parse([
            "doot", "simple task", "-all"
                               ],
            doot_specs=[], cmds={"run": cmd_mock}, tasks={"simple task": task_mock},
            )
        assert(result.on_fail(False).head.name() == "doot")
        assert(result.on_fail(False).cmd.name() == "run")
        assert(bool(result.on_fail(False).tasks['simple task']()))
        assert(result.on_fail(False).tasks['simple task'].all() == True)

    def test_task_args_default(self, ctor, mocker):
        cmd_mock                   = mock_parse_cmd()
        task_mock                  = mock_parse_task(params=[{"name":"all"}])

        parser = ctor()
        result = parser.parse([
            "doot", "list", # no "-all"
                               ],
            doot_specs=[], cmds={"run": cmd_mock}, tasks={"list": task_mock},
            )
        assert(result.on_fail(False).head.name() == "doot")
        assert(bool(result.on_fail(False).tasks.list()))
        assert(result.on_fail(False).tasks.list.all() == False)

    def test_tasks_dup_fail(self, ctor, mocker):
        cmd_mock                   = mock_parse_cmd()
        task_mock                  = mock_parse_task(params=[{"name":"all"}])

        parser = ctor()
        with pytest.raises(doot.errors.DootParseError):
            parser.parse([
                "doot", "list", "-all", "list"
                         ],
                doot_specs=[], cmds={"run": cmd_mock}, tasks={"list": task_mock}
            )

    def test_positional_cmd_arg(self, ctor, mocker):
        cmd_mock                   = mock_parse_cmd(params=[DootParamSpec("test", type=str, positional=True)])
        task_mock                  = mock_parse_task(params=[{"name":"key"}])

        parser = ctor()
        result = parser.parse([
            "doot", "example", "blah"
                        ],
            doot_specs=[], cmds={"example": cmd_mock}, tasks={"other": task_mock},
            )
        assert(result.cmd.name == "example")
        assert(result.cmd.args.test == "blah")

    def test_positional_cmd_arg_seq(self, ctor, mocker):
        cmd_mock                   = mock_parse_cmd(params=[
            DootParamSpec(name="first", type=str, positional=True),
            DootParamSpec(name="second", type=str, positional=True)
            ])
        task_mock                  = mock_parse_task(params=[{"name":"key"}])

        parser = ctor()
        result = parser.parse([
            "doot", "example", "blah", "bloo"
                        ],
            doot_specs=[], cmds={"example": cmd_mock}, tasks={}
            )
        assert(result.cmd.args.first == "blah")
        assert(result.cmd.args.second == "bloo")

    def test_positional_task_arg(self, ctor, mocker):
        cmd_mock                   = mock_parse_cmd()
        task_mock                  = mock_parse_task(params=[{"name":"test", "type":str, "positional":True, "default":""}])

        parser = ctor()
        result = parser.parse([
            "doot", "example", "blah"
                        ],
            doot_specs=[], cmds={"run": cmd_mock}, tasks={"example": task_mock},
            )
        assert("example" in result.tasks)
        assert(result.tasks.example.test == "blah")

    def test_positional_taskarg_seq(self, ctor, mocker):
        cmd_mock                   = mock_parse_cmd()
        task_mock                  = mock_parse_task(params=[
            {"name":"first", "type":str, "positional":True},
            {"name":"second", "type":str, "positional":True}
            ])

        parser = ctor()
        result = parser.parse([
            "doot", "example", "blah", "bloo"
                        ],
            doot_specs=[], cmds={"run": cmd_mock}, tasks={"example": task_mock},
            )
        assert(result.tasks.example.first == "blah")
        assert(result.tasks.example.second == "bloo")

    def test_simple_head_arg(self, ctor, mocker):
        param = DootParamSpec("key", bool)
        parser   = ctor()
        result   = parser.parse(["doot", "-key"],
            doot_specs=[param], cmds={}, tasks={}
            )

        assert(result.head.args.key is True)

    def test_simple_short_arg(self, ctor, mocker):
        param = DootParamSpec("key", bool)
        parser   = ctor()
        result   = parser.parse(["doot", "-k"],
            doot_specs=[param], cmds={}, tasks={}
            )

        assert(result.head.args.key is True)

    def test_simple_invert(self, ctor, mocker):
        param = DootParamSpec("key", bool)
        parser   = ctor()
        result   = parser.parse(["doot", "-no-key"],
            doot_specs=[param], cmds={}, tasks={}
            )

        assert(result.head.args.key is False)

    def test_simple_multi(self, ctor, mocker):
        param    = DootParamSpec("key", bool)
        param2   = DootParamSpec("other", bool)
        parser   = ctor()
        result   = parser.parse(["doot", "-no-key", "-other"],
            doot_specs=[param, param2], cmds={}, tasks={}
            )

        assert(result.head.args.key is False)
        assert(result.head.args.other is True)

    def test_simple_default(self, ctor, mocker):
        param    = DootParamSpec("key", bool)
        parser   = ctor()
        result   = parser.parse(["doot"],
            doot_specs=[param], cmds={}, tasks={}
            )

        assert(result.head.args.key is False)

    def test_simple_assign(self, ctor, mocker):
        param    = DootParamSpec("key", str, prefix="--")
        parser   = ctor()
        result   = parser.parse(["doot", "--key=blah"],
            doot_specs=[param], cmds={}, tasks={}
            )

        assert(result.head.args.key == "blah")

    def test_assign_fail_with_wrong_prefix(self, ctor, mocker):
        param    = DootParamSpec("key", str, prefix="-")
        parser   = ctor()

        with pytest.raises(doot.errors.DootParseError):
            parser.parse(["doot", "--key=blah"],
                doot_specs=[param], cmds={}, tasks={}
                )

    def test_simple_follow_assign(self, ctor, mocker):
        param    = DootParamSpec("key", str)
        parser   = ctor()
        result   = parser.parse(["doot", "-key", "blah"],
            doot_specs=[param], cmds={}, tasks={}
            )

        assert(result.head.args.key == "blah")

    def test_simple_prefix_change(self, ctor, mocker):
        param    = DootParamSpec("key", str, prefix="--")
        parser   = ctor()
        result   = parser.parse(["doot", "--key", "blah"],
            doot_specs=[param], cmds={}, tasks={}
            )

        assert(result.head.args.key == "blah")

    def test_simple_separator_change(self, ctor, mocker):
        param    = DootParamSpec("key", str, separator="%%", prefix="--")
        parser   = ctor()
        result   = parser.parse(["doot", "--key%%blah"],
            doot_specs=[param], cmds={}, tasks={}
            )

        assert(result.head.args.key == "blah")

    def test_simple_cmd(self, ctor, mocker):
        cmd_mock = mock_parse_cmd(params=[DootParamSpec(name="key")])
        parser   = ctor()
        result   = parser.parse(["doot", "list"],
            doot_specs=[], cmds={"list" : cmd_mock}, tasks={}
            )

        assert(result.cmd.name == "list")
        assert(result.cmd.args.key is False)

    def test_simple_cmd_arg(self, ctor, mocker):
        cmd_mock            = mocker.MagicMock()
        type(cmd_mock).param_specs = mocker.PropertyMock(return_value=[
            DootParamSpec(name="key")
            ])
        param = DootParamSpec("key", bool)
        parser   = ctor()
        result   = parser.parse(["doot", "list", '-key'],
            doot_specs=[], cmds={"list" : cmd_mock}, tasks={}
            )

        assert(result.cmd.name == "list")
        assert(result.cmd.args.key is True)

    def test_cmd_default(self, ctor, mocker):
        cmd_mock                   = mock_parse_cmd()
        task_mock                  = mock_parse_task(params=[{"name":"key"}])

        parser   = ctor()
        result   = parser.parse(["doot" , "val"],
            doot_specs=[], cmds={"run": cmd_mock }, tasks={"val": task_mock}
            )

        assert(result.cmd.name == "run")

    def test_simple_task(self, ctor, mocker):
        cmd_mock             = mock_parse_cmd()
        task_mock            = mock_parse_task(params=[{"name":"key"}])

        parser               = ctor()
        result               = parser.parse(["doot", "list", '-key'],
            doot_specs=[], cmds={"run": cmd_mock }, tasks={"list" : task_mock}
            )

        assert(result.cmd.name == "run")
        assert("list" in result.tasks)
        assert(result.tasks.list.key is True)

    def test_simple_task_sequence(self, ctor, mocker):
        cmd_mock   = mock_parse_cmd()
        task_mock  = mock_parse_task(params=[{"name":"key", "type":bool}])
        task_mock2 = mock_parse_task(params=[{"name":"other", "type":bool}])

        parser     = ctor()
        result     = parser.parse(["doot", "list", "-key", "blah", "-other"],
            doot_specs=[], cmds={"run": cmd_mock }, tasks={"list" : task_mock, "blah": task_mock2}
            )

        assert(result.cmd.name == "run")
        assert("list" in result.tasks)
        assert(result.tasks.list.key is True)
        assert("blah" in result.tasks)
        assert(result.tasks.blah.other is True)



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import scrapy
from scrapy import signals
from scrapy.utils.log import log_scrapy_info
from itemadapter import is_item, ItemAdapter
from scrapy.utils.httpobj import urlparse_cached
import gzip
import pickle
from time import time
from w3lib.http import headers_dict_to_raw, headers_raw_to_dict

import scrapy
from scrapy.http import Headers, Response
from scrapy.http.request import Request
from scrapy.responsetypes import responsetypes
from scrapy.spiders import Spider
from scrapy.utils.httpobj import urlparse_cached
from itemadapter import ItemAdapter
from scrapy.exporters import XmlItemExporter
from scrapy.crawler import CrawlerProcess

class SimpleFilter(logmod.Filter):

    def filter(self, record):
        record.args = {x:y.replace("\n"," ") for x,y in record.args.items()}
        record.msg  = record.msg.replace("\n", " ")
        return True

class CrawlerProcessFix(CrawlerProcess):
    """
    Modified scrapy.crawler.CrawlerProcess
    that doesn't call `configure_logging`, and so doesnt mess up
    existing loggers
    """

    def __init__(self, settings=None, install_root_handler=False):
        if not install_root_handler:
            super(CrawlerProcess, self).__init__(settings)
        else:
            logging.warning("Using standard scrapy crawler, will override log settings")
            super().__init__(settings, install_root_handler=install_root_handler)

        # skipped -> configure_logging(self.settings, install_root_handler)
        # log_scrapy_info(self.settings)
        logging.debug("CrawlerProcess Settings: %s", list(self.settings._to_dict().keys()))
        self._initialized_reactor = False
        filter_nls = SimpleFilter()
        for logName in [f"scrapy.{x}" for x in dir(scrapy) if "__" not in x]:
            logmod.getLogger(logName).addFilter(filter_nls)
        logmod.getLogger("scrapy.statscollector").addFilter(filter_nls)



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import tracemalloc
import linecache
import fnmatch

class JGDVMalloc:
    """ see https://docs.python.org/3/library/tracemalloc.html

    example:
    with JGDVMalloc(2) as dm:
        dm.whitelist(__file__)
        dm.blacklist("*tracemalloc.py", all_frames=False)
        val = 2
        dm.snapshot("simple")
        vals = [random.random() for x in range(1000)]
        dm.current()
        dm.snapshot("list")
        vals = None
        dm.current()
        dm.snapshot("cleared")

    dm.compare("simple", "list")
    dm.compare("list", "cleared")
    dm.compare("list", "simple")
    dm.inspect("list")
    """

    def __init__(self, num_frames=5):
        self.num_frames      = num_frames
        self.started         = False
        self.snapshots       = []
        self.named_snapshots = {}
        self.filters         = []
        self.human = lambda num, sign=False: tracemalloc._format_size(num, sign)

    def __enter__(self):
        tracemalloc.start(self.num_frames)
        self.started = True
        self.snapshot(name="init")
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.snapshot(name="final")
        self.started = False
        print(f"Recorded {len(self.snapshots)} snapshots")
        return

    def whitelist(self, file_pat:str, lineno:int=None, all_frames=True):
        self.filters.append(
            tracemalloc.Filter(True,
                               filename_pattern=file_pat,
                               lineno=lineno,
                               all_frames=all_frames)
            )

    def blacklist(self, file_pat:str, lineno:int=None, all_frames=True):
        self.filters.append(
            tracemalloc.Filter(False,
                               filename_pattern=file_pat,
                               lineno=lineno,
                               all_frames=all_frames)
            )

    def check_file_pat(self, file_pat, file_name) -> bool:
        return fnmatch.fnmatch(file_name, file_name)

    def snapshot(self, name=None):
        if not self.started:
            raise RuntimeError("DebugMalloc needs to have been entered")

        snap = tracemalloc.take_snapshot()
        self.snapshots.append(snap)
        if name and name not in self.named_snapshots:
            self.named_snapshots[name] = snap

        tracemalloc.clear_traces()

    def get_snapshot(self, val):
        match val:
            case int() if 0 <= val < len(self.snapshots):
                snap = self.snapshots[val]
            case int() if val < 0:
                snap = self.snapshots[val]
            case str() if val in self.named_snapshots:
                snap = self.named_snapshots[val]
            case _:
                raise TypeError(val)

        return snap.filter_traces(self.filters)


    def current(self, val=None):
        traced = tracemalloc.get_traced_memory()
        print("-- Current Memory: {:<15}, Peak: {:<15}".format( self.human(traced[0]), self.human(traced[1])))
        if val:
            print("Value allocated at: ", tracemalloc.get_object_traceback(val))

    def inspect(self, val, type="traceback"):
        print(f"\n-- Inspecting {val}")
        snap = self.get_snapshot(val)
        for stat in snap.statistics(type):
            self._print_stat(stat)

    def compare(self, val1, val2, type="traceback"):
        print(f"\n-- Comparing {val1} to {val2}")
        snap1 = self.get_snapshot(val1)
        snap2 = self.get_snapshot(val2)

        differences = snap2.compare_to(snap1, type)
        diff_count  = len(differences)
        for i, stat in enumerate(differences):
            print(f"- {val1} -> {val2} diff {i}/{diff_count}:")
            self._print_diff(stat)

    def _print_diff(self, stat):
        """ Print a Trace memory comparison """
        assert(isinstance(stat, tracemalloc.StatisticDiff))
        print(f"+-: {self.human(stat.size_diff, True)}, {stat.count_diff} blocks")
        self._print_stat(stat)

    def _print_stat(self, stat):
        """ Print a Traced memory snapshot """
        assert(isinstance(stat, (tracemalloc.Statistic, tracemalloc.StatisticDiff)))
        tb = stat.traceback
        print(f"Frame/Count/Size: {len(stat.traceback):}, {stat.count:}, {self.human(stat.size)}")
        for x in range(len(tb) -1):
            frame = tb[x]
            path = pl.Path(frame.filename)
            print(f"  {path.name:<15}", "line:", f"{frame.lineno:<7}: {linecache.getline(frame.filename, frame.lineno).strip()}")

        frame = tb[-1]
        path = pl.Path(frame.filename)
        print(f"* {path.name:<15}", "line:", f"{frame.lineno:<7}: {linecache.getline(frame.filename, frame.lineno).strip()}")
        print()



"""
A Pattern collects events together and cycles them
"""
##-- imports
from __future__ import annotations

from .pattern_iterator import PatternIterator
from .utils import TIME_T
from fractions import Fraction, gcd
from functools import reduce
from math import floor

import logging as root_logger
##-- end imports


logging = root_logger.getLogger(__name__)

class Pattern:
    """ A Collection of Events """

    @staticmethod
    def lcm(a,b):
        """ Get the lowest common multiple for two fractions """
        x = abs(a)
        gcd_result = gcd(a,b)
        x_prime = x / gcd_result
        y = abs(b)
        result = x_prime * y
        return result

    def __init__(self, a, vals=None):
        self.arc        = a.copy()
        # components :: [ Event || Pattern ]
        self.components = sorted(vals or [], key=lambda x: x.key())
        self.time_type  = TIME_T.CLOCK

    # TODO: add a state for random generator
    def __call__(self, count, just_values=False, rand_s=None):
        """
        Query the Pattern for a given time
        """
        pattern_range   = self.arc.size()
        f_count         = floor(count)
        position        = count - (f_count * (f_count >= pattern_range))
        scaled_position = position / pattern_range
        results         = []
        # TODO: this could probably be better as a binary tree / beachline
        for x in self.components:
            results += x(scaled_position)

        if just_values:
            results = [x.values for x in results]

        return results


    def __contains__(self, other):
        """ Test whether a given object or time is within this patterns bounds """
        return other in self.arc


    def __repr__(self):
        base_count = self.denominator()
        collection = []
        for y in range(base_count):
            q = self(Fraction(y, base_count), True)
            collection.append(q)

        most_simultaneous = max([len(x) for x in collection])
        rows = [[] for x in range(most_simultaneous)]
        output = "\n|" + ("-" * base_count) + "|\n"
        for x in collection:
            len_x = len(x)
            for j,r in enumerate(rows):
                if j < len_x:
                    r.append(x[j])
                else:
                    r.append('*')

        for row in rows:
            output += "|{}|\n".format("".join(row))

        return output

    def __str__(self):
        """ Print in the same format the parser reads """
        return repr(self)


    def __add__(self, other):
        """ Concatenates two patterns together """
        l_comps = [self]
        r_comps = [other]
        if isinstance(self, PatternSeq):
            l_comps = self.components
        if isinstance(other, PatternSeq):
            r_comps = other.components

        # TODO: is this arc correct?
        return PatternSeq(self.arc,
                          l_comps + r_comps)

    def __sub__(self, other):
        """ Remove a pattern from another
        Equivalent to formatting the pattern with silence
        """
        return None

    def __mul__(self, other):
        """ Stacks two patterns together """
        l_comps = [self]
        r_comps = [other]
        if isinstance(self, PatternSeq):
            l_comps = self.components
        if isinstance(other, PatternSeq):
            r_comps = other.components

        return Pattern(self.arc,
                       l_comps + r_comps)

    def key(self):
        """ Key the Pattern by its start time, for sorting """
        return self.arc.start

    def base(self):
        """ Get all used fractions within this arc, scaled appropriately by offset
        and pattern size """
        counts = set()
        pattern_range = self.arc.size()
        for x in self.components:
            counts.update([a * pattern_range for a in x.base()])
        return counts

    def denominator(self):
        #TODO: use https://stackoverflow.com/questions/49981286/
        base_count = reduce(gcd, self.base(), 2).denominator
        return base_count

    def iter(self, just_values=True, rand_s=None):
        """ Treat the pattern as an iterator """
        return PatternIterator(self, just_values=just_values, rand_s=rand_s)

    def format(self, a_dict):
        """ Apply a substitution dictionary to variables in the pattern """
        raise NotImplementedError()

    def copy(self, deep=False):
        """ Copy the pattern for modification """
        raise NotImplementedError()

    def apply_to(self, other):
        """ Combine two patterns, using the structure of left one """
        raise NotImplementedError()

class PatternSeq(Pattern):

    def __call__(self, count, just_values=False, rand_s=None):
        """
        Query the Pattern for a given time
        """
        f_count = floor(count)
        mod_f = f_count % len(self.components)
        return self.components[mod_f](count, just_values)

class PatternChoice(Pattern):
    def __call__(self, count, just_values=False, rnd_s=None):
        """
        When called chooses from one of the options,
        using the random_state
        """
        raise NotImplementedError()

class PatternOptional(Pattern):
    def __call__(self, count, just_values=False, rnd_s=None):
        """
        When called, either returns nothing, or
        calcs the components as normal
        """
        #if rand_s:
        # TODO: make call take a dict,
        # and return ([], {})?
        ## return []
        # else:
        # return super().__call__ ...
        return []



"""
Provides default comparison functions to be used in the RBTree
Including standard <, > and ==,
along with arc comparisons for beachline use
"""
#pylint: disable=unused-argument
from enum import Enum
import logging as root_logger

logging = root_logger.getLogger(__name__)
Directions = Enum('Directions', 'LEFT RIGHT')

#funcs take a node, a value, and data to contextualise

def default_comparison(a, b, comp_data):
    """ Standard smallest to largest comparison function """
    if a.value < b:
        return Directions.RIGHT
    return Directions.LEFT

def inverted_comparison(a, b, comp_data):
    """ Standard largest to smallest comparison function """
    if a.value < b:
        return Directions.LEFT
    return Directions.RIGHT

def default_equality(a, b, eq_data):
    """ Standard Equality test """
    return a.value == b

def arc_equality(a, b, eq_data):
    """ Compare two arcs for equality """
    #return true if  a.pred|a < b < a|a.succ
    l_inter, r_inter = __arc_intersects(a, b, eq_data)
    result = False
    if l_inter is not None and r_inter is not None:
        result = l_inter < b < r_inter
    elif r_inter is not None:
        result = b < r_inter
    elif l_inter is not None:
        result = l_inter < b
    return result

def arc_comparison(a, b, comp_data):
    """ Function to compare an arc and xposition
    Used in Beachline/Voronoi """
    l_inter, r_inter = __arc_intersects(a, b, comp_data)
    pred_self = False
    self_succ = False

    if l_inter is None and r_inter is None: #Base case: single arc
        if b < a.value.fx:
            return Directions.LEFT
        else:
            return Directions.RIGHT

    if l_inter is not None:
        if b < l_inter:
            pred_self = True
    if r_inter is not None:
        if r_inter < b:
            self_succ = True

    if pred_self:
        return Directions.LEFT
    if self_succ:
        return Directions.RIGHT

    return Directions.LEFT

def __arc_intersects(a, b, comp_data):
    """ Internal function to test if two arcs intersect """
    pred = a.getPredecessor()
    succ = a.getSuccessor()
    pred_intersect = None
    succ_intersect = None
    pred_intersect_out = None
    succ_intersect_out = None
    pred_above_self = None
    succ_above_self = None

    if pred is not None:
        pred_intersect = a.value.intersect(pred.value)
        pred_above_self = a.value.fy < pred.value.fy

    if succ is not None:
        succ_intersect = succ.value.intersect(a.value)
        succ_above_self = a.value.fy < succ.value.fy

    if pred_intersect is not None and len(pred_intersect) == 1:
        pred_intersect_out = pred_intersect[0, 0]
    elif pred_intersect is not None and len(pred_intersect) == 2:
        if pred_above_self:
            pred_intersect_out = pred_intersect[0, 0]
        else:
            pred_intersect_out = pred_intersect[1, 0]

    if succ_intersect is not None and len(succ_intersect) == 1:
        succ_intersect_out = succ_intersect[0, 0]
    elif succ_intersect is not None and len(succ_intersect) == 2:
        if succ_above_self:
            succ_intersect_out = succ_intersect[1, 0]
        else:
            succ_intersect_out = succ_intersect[0, 0]

    return (pred_intersect_out, succ_intersect_out)



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import gzip
import pickle
from time import time

import scrapy
from itemadapter import ItemAdapter, is_item
from scrapy import signals
from scrapy.crawler import CrawlerProcess
from scrapy.exporters import XmlItemExporter
from scrapy.http import Headers, Response
from scrapy.http.request import Request
from scrapy.responsetypes import responsetypes
from scrapy.spiders import Spider
from scrapy.utils.httpobj import urlparse_cached
from w3lib.http import headers_dict_to_raw, headers_raw_to_dict

class JGDVBasicSpider(scrapy.Spider):
    """
    Basic Doot Spider that stores locs and its parent task
    """

    def __init__(self, name=None, locs=None, urls=None, domains=None, task=None):
        super().__init__(name)
        self.locs            = locs
        self.task            = task
        self.start_urls      = urls
        self.allowed_domains = domains or []

    def parse(self, response):
        page     = response.url.split("/")[-2]
        filename = self.locs.data / f"{page}.html"
        pl.Path(filename).write_bytes(response.body)
        yield { "data": response.url }



#!/usr/bin/env python3
"""
Utility class for Twitter Archive Automation,
reads a todo file, ready for downloading, org translation and tagging
"""

##-- imports
from __future__ import annotations

import pathlib as pl
import re
from collections import defaultdict
from dataclasses import InitVar, dataclass, field
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeAlias, TypeVar, Union, cast)

##-- end imports

TagSet : TypeAlias = set[str]
ID     : TypeAlias = str

tag_map : Final[Callable] = lambda: defaultdict(list)

@dataclass
class TweetTodoFile:
    """
    Read a file of tweet references and their tags

    """
    tags_mapping : dict[ID, TagSet] = field(default_factory=tag_map)
    sep          : str              = field(default="_:_")
    ext          : str              = field(default=".tweets")
    remainder    : List[str]        = field(default_factory=list)

    @staticmethod
    def read(p:pl.Path, id_regex=r"status/(\d+)\??"):
        obj   = TweetTodoFile()
        reg   = re.compile(id_regex)
        with open(p, 'r') as f:
            lines = f.readlines()

        for line in lines:
            try:
                url, tags         = line.split(obj.sep)
                id_n              = reg.search(url)[1]
                obj.tags_mapping[id_n].update(tags.strip().split(","))
            except:
                obj.remainder.append(line)

        return obj

    def __len__(self):
        return len(self.tags_mapping)

    def __bool__(self):
        return bool(self.tags_mapping)

    def ids(self) -> Set[ID]:
        return set(self.tags_mapping.keys())



#!/usr/bin/env python3
"""
A Runner that uses twisted's Reactor api
"""
##-- imports
from __future__ import annotations

# import abc
# import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Self)
# from uuid import UUID, uuid1
# from weakref import ref

# from bs4 import BeautifulSoup
# import boltons
# import construct as C
# import dirty-equals as deq
# import graphviz
# import matplotlib.pyplot as plt
# import more_itertools as itzplus
import networkx as nx
# import numpy as np
# import pandas
# import pomegranate as pom
# import pony import orm
# import pronouncing
# import pyparsing as pp
# import rich
# import seaborn as sns
# import sklearn
# import stackprinter # stackprinter.set_excepthook(style='darkbg2')
# import sty
# import sympy
# import tomllib
# import toolz
# import tqdm
# import validators
# import z3
# import spacy # nlp = spacy.load("en_core_web_sm")

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

printer = logmod.getLogger("doot._printer")

from collections import defaultdict
import jgdv
from jgdv.enums.task_respones import TaskResponseEnum as ActRE

from zope.interface import implementer
from twisted.internet import reactor, protocol, defer, threads
from twisted.internet.interfaces import IStreamServerEndpoint

from scrapy.crawler import CrawlerRunner
import scrapy

dry_run                    = doot.args.on_fail(False).cmd.args.dry_run()
head_level    : Final[str] = doot.constants.printer.DEFAULT_HEAD_LEVEL
build_level   : Final[str] = doot.constants.printer.DEFAULT_BUILD_LEVEL
action_level  : Final[str] = doot.constants.printer.DEFAULT_ACTION_LEVEL
sleep_level   : Final[str] = doot.constants.printer.DEFAULT_SLEEP_LEVEL
execute_level : Final[str] = doot.constants.printer.DEFAULT_EXECUTE_LEVEL

reactor_timeout = doot.config.on_fail(2, int).settings.tasks.reactor_timeout()

@jgdv.check_protocol
class JGDVReactorRunner(BaseRunner, TaskRunner_i):
    """ The simplest reactor task runner
      https://stackoverflow.com/questions/8532131

     wraps steps of running jobs/tasks/actions into deferreds
    """

    def __init__(self:Self, *, tracker:TaskTracker_i, reporter:Reporter_i, policy=None):
        super().__init__(tracker=tracker, reporter=reporter, policy=policy)
        scrapy_settings           = dict(doot.config.scrapy.settings.items())
        self.crawler              = CrawlerRunner(settings=scrapy_settings)
        self.reactor              = reactor
        self._enter_msg = "---------- Reactor Starting ----------"
        self._exit_msg  = "---------- Reactor Finished ----------"

    def __call__(self, *tasks:str):
        """ Queues _run_next_task and starts the reactor """
        self.tracker.queue_task(*tasks)
        reactor.callWhenRunning(self._run_next_task)
        reactor.run()

    def _run_next_task(self):
        d = threads.deferToThread(self.tracker.next_for)
        d.pause()
        d.addCallback(self._handle_task)
        d.addCallback(self._handle_task_success)
        d.addErrback(self._handle_failure)

        d.addCallback(self._schedule_next)
        d.unpause()

    def _handle_task(self, task) -> defer.Deferred[None|Task]:
        printer.setLevel("INFO")
        match task:
            case None:
                logging.debug("Task is None")
                return None
            case Job_i():
                self.step += 1
                return defer.maybeDeferred(self._expand_job, task)
            case Task_i():
                self.step += 1
                return defer.maybeDeferred(self._execute_task, task)
            case _:
                raise doot.errors.DootTaskError("Unknown Task Base: %s", task, task=task)

    def _handle_failure(self, failure) -> defer.Failure:
        match failure:
            case doot.errors.DootTaskInterrupt():
                breakpoint()
                return failure
            case doot.errors.DootTaskTrackingError() as err:
                return err.task
            case doot.errors.DootTaskFailed() as err:
                printer.warning("Task Failed: %s : %s", err.task.name, err)
                self.tracker.update_state(err.task, self.tracker.state_e.FAILED)
                return err.task
            case doot.errors.DootTaskError() as err:
                printer.warning("Task Error : %s : %s", err.task.name, err)
                self.tracker.update_state(err.task, self.tracker.state_e.FAILED)
                return err.task
            case doot.errors.DootError() as err:
                printer.warning("Doot Error : %s", err)
                return failure
            case _:
                return failure

    def _schedule_next(self, task) -> defer.Deferred:
        """ Schedules the next available task """
        if task is None and not bool(self.tracker.task_queue):
            printer.info("Scheduling Reactor shutdown")
            return self.reactor.callLater(reactor_timeout, self.reactor.stop)

        if task is None:
            printer.info("Task is none, tracker is: %s", bool(self.tracker.task_queue))
            return self.reactor.callLater(reactor_timeout, self.reactor.stop)

        with logctx(task.spec.print_levels.on_fail(sleep_level).sleep()):
            sleep_len = task.spec.extra.on_fail(self.default_SLEEP_LENGTH, int|float).sleep()
            printer.info("[Sleeping (%s)...]", sleep_len, extra={"colour":"white"})
            return self.reactor.callLater(sleep_len, self._run_next_task)


    def _expand_job(self, job:Job_i) -> Job_i:
        """ turn a job into all of its tasks, including teardowns """
        logmod.debug("-- Expanding Job %s: %s", self.step, job.name)
        with logctx(job.spec.print_levels.on_fail(head_level).head()) as p:
            p.info("---- Job %s: %s", self.step, job.name, extra={"colour":"magenta"})
            if bool(job.spec.actions):
                p.warning("-- Job %s: Actions were found in job spec, but jobs don't _run_next_task actions")

        with logctx(job.spec.print_levels.on_fail(build_level).build()):
            count = 0
            for task in job.build():
                match task:
                    case Job_i():
                        printer.warning("Jobs probably shouldn't build jobs: %s : %s", str(job.name), str(task.name))
                        self.tracker.add_task(task, no_root_connection=True)
                        self.tracker.queue_task(task.name)
                    case Task_i():
                        self.tracker.add_task(task, no_root_connection=True)
                        self.tracker.queue_task(task.name)
                    case DootTaskSpec():
                        self.tracker.add_task(task, no_root_connection=True)
                        self.tracker.queue_task(task.name)
                    case _:
                        raise doot.errors.DootTaskError("Job %s Built a Bad Value: %s", str(job.name), task, task=str(job.name))

                count += 1

        logmod.debug("-- Job %s Expansion produced: %s tasks", job.name, count)
        return job

    def _execute_task(self, task:Task_i) -> defer.Deferred[Task]:
        """ execute a single task's actions """
        with logctx(task.spec.print_levels.on_fail(head_level).head()) as p:
            p.info("---- Task %s: %s", self.step, task.name, extra={"colour":"magenta"})

        d = defer.succeed(task)
        d.pause()
        action_count = 0
        action_result = ActRE.SUCCESS
        with logctx(task.spec.print_levels.on_fail(build_level).build()) as p:
            for action in task.actions:
                match action:
                    case DootActionSpec() if action.fun is None:
                        raise doot.errors.DootTaskError("Task %s Failed: Produced an action with no callable: %s", task.name, action, task=task.spec)
                    case DootActionSpec():
                        match self._execute_action(action_count, action, task):
                            case ActRE.SKIP:
                                action_result = ActRE.SKIP
                            case defer.Deferred() as x:
                                d.addCallback(lambda _: x)
                            case _:
                                pass
                    case _:
                        raise doot.errors.DootTaskError("Task %s Failed: Produced a bad action: %s", task.name, action, task=task.spec)

                action_count += 1
                if action_result is ActRE.SKIP:
                    p.info("------ Remaining Task Actions skipped by Action Instruction")
                    break

            else: # Only try to add more tasks if the actions completed successfully, and weren't skipped
                for new_task in task.maybe_more_tasks():
                    p.warning("Skipping task sub: %s", new_task)

            p.debug("------ Task Executed %s Actions", action_count)
            d.addCallback(lambda _: task)
            d.unpause()
        return d

    def _execute_action(self, count, action, task) -> ActRE|defer.Deferred:
        """ Run the given action of a specific task  """
        if dry_run:
            logging.info("Dry Run: Not executing action: %s : %s", task.name, action, extra={"colour":"cyan"})
            return ActRE.SUCCESS

        logmod.debug("------ Executing Action %s: %s for %s", count, action, task.name)
        result = None
        with logctx(task.spec.print_levels.on_fail(action_level).action()) as p:
            task.state['_action_step'] = count
            p.info("------ Action %s.%s: %s", self.step, count, action.do, extra={"colour":"cyan"})
            p.debug("------ Action %s.%s: args=%s kwargs=%s. state keys = %s", self.step, count, action.args, dict(action.kwargs), list(task.state.keys()))
            action.verify(task.state)
            task.state['_reactor'] = self.reactor
            task.state['_crawler'] = self.crawler

            with logctx(task.spec.print_levels.on_fail(execute_level).execute()):
                result = action(task.state)

            p.debug("-- Action Result: %s", result)

        match result:
            case defer.Deferred():
                pass
            case ActRE.SKIP:
                pass
            case None | True:
                result = ActRE.SUCCESS
            case dict():
                task.state.update(result)
                result = ActRE.SUCCESS
            case False | ActRE.FAIL:
                raise doot.errors.DootTaskFailed("Task %s Action Failed: %s", task.name, action, task=task.spec)
            case _:
                raise doot.errors.DootTaskError("Task %s Action %s Failed: Returned an unplanned for value: %s", task.name, action, result, task=task.spec)

        action.verify_out(task.state)

        logmod.debug("------ Action Execution Complete: %s for %s", action, task.name)
        return result



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
printer = logmod.getLogger("doot._printer")
##-- end logging

import decorator
import abc
import string
from tomlguard import TomlGuard

KEY_PATTERN                                = doot.constants.patterns.KEY_PATTERN
MAX_KEY_EXPANSIONS                         = doot.constants.patterns.MAX_KEY_EXPANSIONS
STATE_TASK_NAME_K                          = doot.constants.patterns.STATE_TASK_NAME_K

PATTERN        : Final[re.Pattern]         = re.compile(KEY_PATTERN)
FAIL_PATTERN   : Final[re.Pattern]         = re.compile("[^a-zA-Z_{}/0-9-]")
KEYS_HANDLED   : Final[str]                = "_doot_keys_handler"
ORIG_ARGS      : Final[str]                = "_doot_orig_args"
KEY_ANNOTS     : Final[str]                = "_doot_keys"
EXPANSION_HINT : Final[str]                = "_doot_expansion_hint"
HELP_HINT      : Final[str]                = "_doot_help_hint"
FUNC_WRAPPED   : Final[str]                = "__wrapped__"

class JGDVBaseKey(abc.ABC):
    """ A shared, non-functional base class for keys and variants like MultiKey.
      Use JGDVBaseKey.build for constructing keys
      build takes an 'exp_hint' kwarg dict, which can specialize the expansion

      DootSimpleKeys are strings, wrapped in {} when used in toml.
      so JGDVBaseKey.build("blah") -> DootSimpleKey("blah") -> DootSimpleKey('blah').form =="{blah}" -> [toml] aValue = "{blah}"

      DootMultiKeys are containers of a string `value`, and a list of SimpleKeys the value contains.
      So JGDVBaseKey.build("{blah}/{bloo}") -> DootMultiKey("{blah}/{bloo}", [DootSimpleKey("blah", DootSimpleKey("bloo")]) -> .form == "{blah}/{bloo}"
    """
    dec   = KWrapper
    kwrap = KWrapper

    @staticmethod
    def build(s:str|JGDVBaseKey|pl.Path|dict, *, strict=False, explicit=False, exp_hint:str|dict=None, help=None) -> JGDVBaseKey:
        """ Make an appropriate JGDVBaseKey based on input value
          Can only create MultiKeys if strict = False,
          if explicit, only keys wrapped in {} are made, everything else is returned untouched
          if strict, then only simple keys can be returned
        """
        # TODO annotate with 'help'
        # TODO store expansion args on build
        match exp_hint:
            case "path":
                is_path = True
            case {"expansion": "path"}:
                is_path = True
            case _:
                is_path = False
        result = s
        match s:
            case { "path": x }:
                result = DootPathMultiKey(x)
                exp_hint = "path"
            case pl.Path():
                result = DootPathMultiKey(s)
                exp_hint = "path"
            case DootSimpleKey() if strict:
                result = s
            case JGDVBaseKey():
                result = s
            case str() if not (s_keys := PATTERN.findall(s)) and not explicit and not is_path:
                result = DootSimpleKey(s)
            case str() if is_path and not bool(s_keys):
                result = DootPathSimpleKey(s)
            case str() if is_path and len(s_keys) == 1 and s_keys[0] == s[1:-1]:
                result = DootPathSimpleKey(s[1:-1])
            case str() if is_path and len(s_keys) > 1:
                result = DootPathMultiKey(s)
            case str() if not s_keys and explicit:
                result = DootNonKey(s)
            case str() if len(s_keys) == 1 and s_keys[0] == s[1:-1]:
                result = DootSimpleKey(s[1:-1])
            case str() if not strict:
                result = DootMultiKey(s)
            case _:
                raise TypeError("Bad Type to build a Doot Key Out of", s)

        if exp_hint is not None:
            result.set_expansion_hint(exp_hint)

        return result

    def set_help(self, help:str):
        setattr(self, HELP_HINT, help)

    def set_expansion_hint(self, etype:str|dict):
        match etype:
            case "str" | "path" | "type" | "redirect" | "redirect_multi":
                setattr(self, EXPANSION_HINT, {"expansion": etype, "kwargs": {}})
            case {"expansion": str(), "kwargs": dict()}:
                setattr(self, EXPANSION_HINT, etype)
            case _:
                raise doot.errors.DootKeyError("Bad Key Expansion Type Declared", self, etype)

    def __call__(self, spec, state):
        """ Expand the key using the registered expansion hint """
        match getattr(self, EXPANSION_HINT, False):
            case False:
                raise doot.errors.DootKeyError("No Default Key Expansion Type Declared", self)
            case {"expansion": "str", "kwargs": kwargs}:
                return self.expand(spec, state, **kwargs)
            case {"expansion": "path", "kwargs": kwargs}:
                return self.to_path(spec, state, **kwargs)
            case {"expansion" : "type", "kwargs" : kwargs}:
                return self.to_type(spec, state, **kwargs)
            case {"expansion": "redirect"}:
                return self.redirect(spec)
            case {"expansion": "redirect_multi"}:
                return self.redirect_multi(spec)
            case {"expansion": "coderef"}:
                return self.to_coderef(spec, state)
            case x:
                raise doot.errors.DootKeyError("Key Called with Bad Key Expansion Type", self, x)

    @property
    def form(self) -> str:
        return str(self)

    @property
    def direct(self):
        return str(self).removesuffix("_")

    @property
    def is_indirect(self) -> bool:
        return False

    def redirect(self, spec=None) -> JGDVBaseKey:
        return self

    def to_path(self, spec=None, state=None, chain:list[JGDVBaseKey]=None, locs:DootLocations=None, on_fail:None|str|pl.Path|JGDVBaseKey=Any, symlinks=False) -> pl.Path:
        """
          Convert a key to an absolute path, using registered locations

          The Process is:
          1) redirect the given key if necessary
          2) Expand each part of the keypath, using DootFormatter
          3) normalize it

          If necessary, a fallback chain, and on_fail value can be provided
        """
        locs                 = locs or doot.locs
        key : pl.Path        = pl.Path(self.redirect(spec).form)

        try:
            expanded         : list       = [DootFormatter.fmt(x, _spec=spec, _state=state, _rec=True, _locs=locs) for x in key.parts]
            expanded_as_path : pl.Path    = pl.Path().joinpath(*expanded) # allows ("a", "b/c") -> "a/b/c"

            if bool(matches:=PATTERN.findall(str(expanded_as_path))):
                raise doot.errors.DootLocationExpansionError("Missing keys on path expansion", matches, self)

            return locs.normalize(expanded_as_path, symlinks=symlinks)

        except doot.errors.DootLocationExpansionError as err:
            if bool(chain):
                return chain[0].to_path(spec, state, chain=chain[1:], on_fail=on_fail, symlinks=symlinks)
            match on_fail:
                case None:
                    return None
                case JGDVBaseKey():
                    return on_fail.to_path(spec, state, symlinks=symlinks)
                case pl.Path() | str():
                    return locs.normalize(pl.Path(on_fail),  symlinks=symlinks)
                case _:
                    raise err

    def within(self, other:str|dict|TomlGuard) -> bool:
        return False

    def basic(self, spec:SpecStruct_p, state, locs=None):
        """ the most basic expansion of a key """
        kwargs = spec.params
        return DootKeyGetter.chained_get(str(self), kwargs, state, locs or doot.locs)

    @abc.abstractmethod
    def to_type(self, spec, state, type_=Any, chain:list[JGDVBaseKey]=None, on_fail=Any, **kwargs) -> Any:
        raise NotImplementedError()

    @abc.abstractmethod
    def expand(self, spec=None, state=None, *, rec=False, insist=False, chain:list[JGDVBaseKey]=None, on_fail=Any, locs:DootLocations=None, **kwargs) -> str:
        pass

    def to_coderef(self, spec:None|SpecStruct_p, state) -> None|DootCodeReference:
        match spec:
            case SpecStruct_p():
                kwargs = spec.params
            case None:
                kwargs = {}

        redir = self.redirect(spec)

        if redir not in kwargs and redir not in state:
            return None
        try:
            expanded = self.expand(spec, state)
            ref = DootCodeReference.build(expanded)
            return ref
        except doot.errors.DootError:
            return None




#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import scrapy
from scrapy import signals
from itemadapter import is_item, ItemAdapter
from scrapy.utils.httpobj import urlparse_cached
import gzip
import pickle
from time import time
from w3lib.http import headers_dict_to_raw, headers_raw_to_dict

import scrapy
from scrapy.http import Headers, Response
from scrapy.http.request import Request
from scrapy.responsetypes import responsetypes
from scrapy.spiders import Spider
from scrapy.utils.httpobj import urlparse_cached
from itemadapter import ItemAdapter
from scrapy.exporters import XmlItemExporter
from scrapy.crawler import CrawlerProcess

class SpiderMiddleware:

    # Not all methods need to be defined. If a method is not defined,
    # scrapy acts as if the spider middleware does not modify the
    # passed objects.

    @classmethod
    def from_crawler(cls, crawler):
        # This method is used by Scrapy to create your spiders.
        s = cls()
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        return s

    def process_spider_input(self, response, spider):
        # Called for each response that goes through the spider
        # middleware and into the spider.

        # Should return None or raise an exception.
        return None

    def process_spider_output(self, response, result, spider):
        # Called with the results returned from the Spider, after
        # it has processed the response.

        # Must return an iterable of Request, or item objects.
        for i in result:
            yield i

    def process_spider_exception(self, response, exception, spider):
        # Called when a spider or process_spider_input() method
        # (from other spider middleware) raises an exception.

        # Should return either None or an iterable of Request or item objects.
        pass

    def process_start_requests(self, start_requests, spider):
        # Called with the start requests of the spider, and works
        # similarly to the process_spider_output() method, except
        # that it doesn’t have a response associated.

        # Must return only requests (not items).
        for r in start_requests:
            yield r

    def spider_opened(self, spider):
        pass

class DownloaderMiddleware:
    # Not all methods need to be defined. If a method is not defined,
    # scrapy acts as if the downloader middleware does not modify the
    # passed objects.

    @classmethod
    def from_crawler(cls, crawler):
        # This method is used by Scrapy to create your spiders.
        s = cls()
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        return s

    def process_request(self, request, spider):
        # Called for each request that goes through the downloader
        # middleware.

        # Must either:
        # - return None: continue processing this request
        # - or return a Response object
        # - or return a Request object
        # - or raise IgnoreRequest: process_exception() methods of
        #   installed downloader middleware will be called
        return None

    def process_response(self, request, response, spider):
        # Called with the response returned from the downloader.

        # Must either;
        # - return a Response object
        # - return a Request object
        # - or raise IgnoreRequest
        return response

    def process_exception(self, request, exception, spider):
        # Called when a download handler or a process_request()
        # (from other downloader middleware) raises an exception.

        # Must either:
        # - return None: continue processing this exception
        # - return a Response object: stops process_exception() chain
        # - return a Request object: stops process_exception() chain
        pass

    def spider_opened(self, spider):
        pass



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class DCELAdder:

    def new_vertex(self, loc, data=None, force=False):
        """ Create a new vertex, or reuse an existing vertex.
        to force a new vertex instead of reusing, set force to True
        """
        assert(isinstance(loc, np.ndarray))
        new_vert = None
        matching_vertices = self.vertex_quad_tree.intersect(Vertex.free_bbox(loc))
        logging.debug("Quad Tree Size: {}".format(self.vertex_quad_tree.countmembers()))
        logging.debug("Query result: {}".format(matching_vertices))
        if bool(matching_vertices) and not force:
            #a good enough vertex exists
            new_vert = matching_vertices.pop()
            if data is not None:
                new_vert.data.update(data)
            logging.debug("Found a matching vertex: {}".format(new_vert))
        else:
            #no matching vertex, add this new one
            new_vert = Vertex(loc, data=data, dcel=self)
            logging.debug("No matching vertex, storing: {}, {}".format(new_vert, new_vert.bbox()))

        assert(new_vert is not None)
        return new_vert

    def new_edge(self, origin_vertex, twin_vertex, face=None, twin_face=None, prev_edge=None, twin_prev=None, next_edge=None, twin_next=None, edata=None, vdata=None):
        """ Create a new half edge pair, after specifying its start and end.
            Can set the faces, and previous edges of the new edge pair.
            Returns the outer edge
        """
        #todo: check for an already existing edge
        assert(origin_vertex is None or isinstance(origin_vertex, Vertex))
        assert(twin_vertex is None or isinstance(twin_vertex, Vertex))
        e1 = HalfEdge(origin_vertex, None, dcel=self)
        e2 = HalfEdge(twin_vertex, e1, dcel=self)
        e1.twin = e2
        #Connect with passed in details
        if face is not None:
            assert(isinstance(face, Face))
            face.add_edge(e1)
        if twin_face is not None:
            assert(isinstance(twin_face, Face))
            twin_face.add_edge(e2)
        if prev_edge is not None:
            assert(isinstance(prev_edge, HalfEdge))
            e1.add_prev(prev_edge)
        if twin_prev is not None:
            assert(isinstance(twin_prev, HalfEdge))
            e2.add_prev(twin_prev)
        if next_edge is not None:
            assert(isinstance(next_edge, HalfEdge))
            e1.add_next(next_edge)
        if twin_next is not None:
            assert(isinstance(twin_next, HalfEdge))
            e2.add_next(twin_next)
        if edata is not None:
            e1.data.update(edata)
            e2.data.update(edata)
        if vdata is not None:
            e1.origin.data.update(vdata)
            e2.origin.data.update(vdata)
        self.half_edges.update([e1, e2])
        logging.debug("Created Edge Pair: {}".format(e1.index))
        logging.debug("Created Edge Pair: {}".format(e2.index))
        return e1

    def new_face(self, site=None, edges=None, verts=None, coords=None, data=None):
        """ Creates a new face to link edges """
        used_list = [edges is not None, verts is not None, coords is not None]
        assert(len([x for x in used_list if x]) < 2)
        if site is None:
            site = np.array([0, 0])
        assert(isinstance(site, np.ndarray))
        new_face = Face(site=site, dcel=self, data=data)
        self.faces.add(new_face)
        #populate the face if applicable:
        coord_hull_gen = False
        if coords is not None:
            assert(isinstance(coords, np.ndarray))
            assert(coords.shape[1] == 2)
            hull_coords = Face.hull_from_coords(coords)
            verts = [self.new_vertex(x) for x in hull_coords]
            coord_hull_gen = True

        if verts is not None:
            if not coord_hull_gen:
                verts, _ = Face.hull_from_vertices(verts)
            edges = []
            for s, e in zip(verts, islice(cycle(verts), 1, None)):
                edges.append(self.new_edge(s, e))

        if edges is not None:
            new_face.add_edges(edges)
            self.link_edges_together(edges, loop=True)

        return new_face


    def create_edge(self, origin, end, edata=None, vdata=None, subdivs=0):
        """ Utility to create two vertices, and put them into a pair of halfedges,
        returning a halfedge
        subdivs specifies number of inner segments to the line"""
        assert(isinstance(origin, np.ndarray))
        assert(isinstance(end, np.ndarray))
        v1 = self.new_vertex(origin)
        v2 = self.new_vertex(end)
        e = self.new_edge(v1, v2)
        if vdata is not None:
            assert(isinstance(vdata, dict))
            v1.data.update(vdata)
            v2.data.update(vdata)
        if edata is not None:
            assert(isinstance(edata, dict))
            e.data.update(edata)
        return e

    def create_path(self, vs, close=False, edata=None, vdata=None):
        """ Create multiple half_edges, that connect to one another.
        With optional path closing """
        assert(isinstance(vs, np.ndarray))
        vertices = vs
        path_verts = zip(vertices, islice(cycle(vertices), 1, None))
        path = []
        for a, b in path_verts:
            if not close and (a == vertices[-1]).all():
                continue
            path.append(self.create_edge(a, b, edata=edata, vdata=vdata))
        return path

    def create_bezier(self, vs, edata=None, vdata=None, single=False):
        """ Takes a list of tuples (len 3 or 4), and creates
        approximation lines, that can be triggered later to
        draw the true bezier shape,
        Bezier Tuple: (Start, cps, End)"""
        assert(isinstance(vs, list))
        assert(all([isinstance(x, tuple) for x in vs]))
        if edata is None:
            edata = {}
        edges = []

        if single:
            #create a single, multi breakpoint line
            first = vs[0]
            last = vs[-1]
            e = self.create_edge(first[0], last[-1], edata=edata, vdata=vdata)
            e.data[EdgeE.BEZIER] = vs
            e.twin.data[EdgeE.NULL] = True
            return [e]


        for v in vs:
            if len(v) == 2 and all([isinstance(x, tuple) for x in v]):
                #is a single edge, with different control points for different
                #directions
                raise Exception("Dual Control Point Edges not yet supported")
            elif len(v) == 3:
                #is a single cp bezier
                a, _, b = v
                edge = self.create_edge(a, b, edata=edata, vdata=vdata)
                edge.data[EdgeE.BEZIER] = [v]
                edge.twin.data[EdgeE.NULL] = True
                edges.append(edge)
            elif len(v) == 4:
                #is a two cp bezier
                a, _, _, b = v
                edge = self.create_edge(a, b, edata=edata, vdata=vdata)
                edge.data[EdgeE.BEZIER] = [v]
                edge.twin.data[EdgeE.NULL] = True
            else:
                raise Exception("Unrecognised bezier type: {}".format(len(v)))

        return edges




    def add_to_current_quad_tree(self, verts):
        """ Add the passed in vertices into the current quad tree """
        assert(all([isinstance(x, Vertex) for x in verts]))
        assert(bool(self.quad_tree_stack))
        for x in verts:
            self.vertex_quad_tree.insert(item=x, bbox=x.bbox())



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class TexSections_e(enum.Enum):
    header  = enum.auto()
    footer  = enum.auto()
    meta    = enum.auto()
    preface = enum.auto()
    body    = enum.auto()


class TexBuilder_i:
    """ Utility class for creating tex files
      accumulates entries non-linearly,
      then writes them to a file in a particular order
    """

    def __init__(self):
        self._current  = []
        self._sections = {}
        pass

    def author(self):
        pass

    def header(self):
        pass

    def footer(self):
        pass

    def compile(self, section) -> str:
        """ compile a section, calling str on each element """
        pass

    def escape(self, s:str):
        return

    def commit(self, section:str|TexEnvironment_i|TexStatement_i):
        """
          commit the current state into a particular section
        """
        pass

    def add(self, *args):
        pass

class TexEnvironment_i:
    """ Utility base class for using tex environments as a context manager,
      adding state non-linearly
    """

    def __init__(self):
        self._current = []
        self._sections = {}
        self._target_section = None
        pass

    def __str__(self):
        pass

    def __enter__(self) -> Any:
        #return self
        return

    def __exit__(self, exc_type, exc_value, exc_traceback) -> bool:
       # return False to reraise errors
        return

    def header(self):
        pass

    def footer(self):
        pass

    def file_header(self):
        pass

class TexStatement_i:
    """ Utility class interface for building a particular type of tex statement """

    def __init__(self):
        self._target_section = None
        pass

    def __str__(self):
        pass



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging


NEIGHBOUR_CONDITION = lambda v, n: n.value.contains_vertex(v)

class LineSweep:

    def __call__(self, edge_set=None):
        #pylint: disable=unused-argument
        self.initialise_data(edge_set=edge_set)
        assert(bool(self.event_list))
        assert(bool(self.discovered))
        assert(not bool(self.results))
        assert(self.sweep_y == inf)

        #Main loop of the Intersection Algorithm
        while bool(self.event_list):
            logging.debug("--------------------")
            logging.debug("Event list remaining: {}".format(len(self.event_list)))
            logging.debug("\n".join([repr(x) for x in self.event_list]))

            curr_vert, curr_edge_list = self.get_next_event()
            logging.debug("Curr Vert: {}".format(curr_vert))
            self.update_sweep(curr_vert)
            self.debug_chain("Initial")

            logging.debug("Searching tree")
            closest_node, _ = self.search_tree(curr_vert)
            upper_set, contain_set, lower_set = self.determine_sets(curr_vert,
                                                                    closest_node,
                                                                    curr_edge_list.copy())
            self.report_intersections(curr_vert, upper_set, contain_set, lower_set)

            #todo: delete non-flat points of the non-flat event
            self.delete_values(contain_set, curr_x=curr_vert.loc[0])
            self.delete_values(lower_set, curr_x=curr_vert.loc[0])
            self.debug_chain("Deleted")

            #insert the segments with the status line a little lower
            candidate_lines = contain_set.union(upper_set)
            new_nodes = self.insert_values(candidate_lines, curr_x=curr_vert.loc[0])

            #Calculate additional events
            self.debug_chain("Inserted")

            self.handle_new_events(curr_vert, new_nodes)

        assert(self.sweep_y != inf)
        assert(not bool(self.event_list))
        return self.results

    def initialise_data(self, edge_set=None):
        """ Setup the internal structures to run a line sweep algorithm """
        logging.debug("Starting line intersection algorithm")
        self.results = []
        #setup the set of edges to intersect
        if edge_set is None:
            self.edge_set = self.dcel.half_edges.copy()
        else:
            assert(isinstance(edge_set, set))
            #get the twins as well
            twins = [x.twin for x in edge_set]
            edge_set.update(twins)
            self.edge_set = edge_set
        assert(self.edge_set is not None)
        self.lower_edges = [e for e in self.edge_set if not e.is_upper()]
        self.event_list = [HeapWrapper(x.origin, x, desc="initial") \
                           for x in self.edge_set.difference(self.lower_edges)]
        self.event_list += [HeapWrapper(x.origin, x.twin, desc="initial_twin") \
                            for x in self.lower_edges]
        heapq.heapify(self.event_list)
        self.discovered.update([x.ordinal for x in self.event_list])

        logging.debug("EdgeSet: {}, lower_edges: {}".format(len(self.edge_set),
                                                            len(self.lower_edges)))
        logging.debug("Event_list: {}".format(len(self.event_list)))

    def determine_sets(self, curr_vert, closest, candidates):
        """ TODO: describe """
        #pylint: disable=no-self-use
        assert(isinstance(curr_vert, Vertex))
        upper_set = set()
        contain_set = set()
        lower_set = set()

        if closest is not None:
            closest_segment = closest.value
            #if a line is horizontal, switch to a call(y) while,
            #and get lines until not in horizontal bounds
            candidate_nodes = closest.get_neighbours_while(partial(NEIGHBOUR_CONDITION, curr_vert))
            candidates += [x.value for x in candidate_nodes]
            candidates.append(closest_segment)

        logging.debug("Candidates: {}".format(candidates))
        #todo: adapt for horizontal lines
        #init a dictionary for intersections for each individual horizontal line
        #foreach individual horizontal line, test candidates until failure, then break
        for c in candidates:
            if c.origin == curr_vert:
                upper_set.add(c)
            elif c.twin.origin == curr_vert:
                lower_set.add(c)
            elif c.contains_vertex(curr_vert):
                contain_set.add(c)
        logging.debug("Upper Set: {}".format([x.index for x in upper_set]))
        logging.debug("Contain Set: {}".format([x.index for x in contain_set]))
        logging.debug("Lower Set: {}".format([x.index for x in lower_set]))
        return (upper_set, contain_set, lower_set)

    def handle_new_events(self, curr_vert, new_nodes):
        """ Given new nodes to add, process them """
        #pylint: disable=unused-argument
        logging.debug("Finding new Events")
        #todo: only find new events if there is a set of non-horizontal lines
        if not bool(new_nodes):
            closest_node, _ = self.search_tree(curr_vert)
            if closest_node is None:
                return
            left_n = closest_node.get_predecessor()
            right_n = closest_node.get_successor()
            if left_n is not None:
                self.find_new_events(left_n.value, closest_node.value, curr_vert.to_array())
            if right_n is not None:
                self.find_new_events(closest_node.value, right_n.value, curr_vert.to_array())

        else:
            logging.debug("New nodes added")
            assert(bool(self.status_tree))
            if not bool(new_nodes):
                return
            #TODO: this could be more efficient
            chain = self.status_tree.get_chain()
            ordinalered = [x for x in chain if x in new_nodes]
            logging.debug("New Nodes: {}".format([x.value.index for x in ordinalered]))
            leftmost = ordinalered[0]
            leftmost_n = leftmost.get_predecessor()
            if leftmost_n is not None and leftmost is not None:
                self.find_new_events(leftmost_n.value,
                                     leftmost.value,
                                     curr_vert.to_array())

            rightmost = ordinalered[-1]
            rightmost_n = rightmost.get_successor()

            if rightmost is not None and rightmost_n is not None:
                self.find_new_events(rightmost.value,
                                     rightmost_n.value,
                                     curr_vert.to_array())

    def find_new_events(self, a, b, loc):
        """ Given two edges and a location of the sweep line,
        find any valid intersection
        """
        assert(isinstance(loc, np.ndarray))
        assert(isinstance(a, HalfEdge))
        assert(isinstance(b, HalfEdge))

        logging.debug("Finding Events for: {}, {}, {}".format(a.index, b.index, loc))
        intersection = a.intersect(b)
        match_vert = None
        if intersection is None:
            logging.debug("No intersection")
            return
        if a.is_flat() or b.is_flat():
            logging.debug("Forcing same y position")
            intersection[1] = loc[1]

        #TODO: only works on cartesian
        if intersection[1] > loc[1]:
            logging.debug("Intersection too high: {} -- {}".format(intersection[1], loc[1]))
            return
        if intersection[1] < loc[1] or\
           (intersection[1] == loc[1] and loc[0] <= intersection[0]):
            logging.debug("Within bounds")
            match_vert = self.dcel.new_vertex(intersection)
            if match_vert in self.discovered:
                logging.debug("Vertex already discovered")
                return
        #finally, success!:
        if match_vert is not None:
            self.discovered.add(match_vert)
            wrapped = HeapWrapper(match_vert, a, desc="newEvent")
            wrapped2 = HeapWrapper(match_vert, b, desc="newEvent")
            logging.debug("Adding: {}".format(wrapped))
            heapq.heappush(self.event_list, wrapped)
            heapq.heappush(self.event_list, wrapped2)

    #------------------------------
    # def UTILITIES
    #------------------------------

    def debug_chain(self, the_str=None):
        """ log the flattened structure of the internal status tree """
        if the_str is None:
            the_str = ""
        chain = [x.value.index for x in self.status_tree.get_chain()]
        logging.debug("{} Tree Chain is: {}".format(the_str, chain))
        encountered = set()
        for x in chain:
            if x in encountered:
                raise Exception("Duplicated: {}".format(x))
            encountered.add(x)

    def report_intersections(self, v, u, c, l):
        """ Add intersection results when applicable """
        #todo: for each horizontal crossing point separately
        if sum([len(x) for x in [u, l, c]]) > 1:
            logging.debug("Reporting intersections")
            self.results.append(IntersectResult(v, u, c, l))

    def get_next_event(self):
        """ Get all segments that are on the current vertex """
        curr_vert, curr_edge_list = pop_while_same(self.event_list)
        assert(not bool(curr_edge_list) or all([isinstance(x, HalfEdge) for x in curr_edge_list]))
        return curr_vert, curr_edge_list

    def delete_values(self, values, curr_x):
        """ Remove a value from the status tree  """
        logging.debug("--------------------")
        logging.debug("Deleting values: {}".format([x.index for x in values]))
        assert(all([x.is_upper() for x in values]))
        chain = [x.value.index for x in self.status_tree.get_chain()]
        logging.debug("Chain: {}".format(chain))
        self.status_tree.delete_value(*values, cmp_data={'y':self.sweep_y, 'nudge': -SWEEP_NUDGE,
                                                         'x': curr_x - D_EPSILON})

        #Verify:
        chain = [x.value.index for x in self.status_tree.get_chain()]
        assert(all([x.index not in chain for x in values]))

    def insert_values(self, candidates, curr_x):
        """ Insert values into the status tree """
        logging.debug("Inserting values: {}".format([x.index for x in candidates]))
        flat_lines = set([x for x in candidates if x.is_flat()])
        #insert only non-flat lines
        to_add = candidates.difference(flat_lines)
        assert(all([x.is_upper() for x in to_add]))
        new_nodes = self.status_tree.insert(*to_add,
                                            cmp_data={'y':self.sweep_y, 'nudge': SWEEP_NUDGE,
                                                      'x': curr_x + D_EPSILON})
        new_nodes += self.status_tree.insert(*flat_lines,
                                             cmp_data={'y':self.sweep_y, 'nudge':SWEEP_NUDGE,
                                                       'x': curr_x + D_EPSILON})

        return new_nodes

    def update_sweep(self, curr):
        """ Increment the position of the frontier """
        assert(isinstance(curr, Vertex))
        candidate = curr.to_array()[1]
        if candidate > self.sweep_y:
            raise Exception("Sweep line moved in wrong direction: {} vs {}".format(candidate,
                                                                                   self.sweep_y))
        self.sweep_y = candidate

    def search_tree(self, curr):
        """ Search the status tree for a vertex """
        assert(isinstance(curr, Vertex))
        closest_node, d = self.status_tree.search(curr.to_array(),
                                                  cmp_data={'y': self.sweep_y,
                                                            'x': curr.to_array()[0]},
                                                  closest=True,
                                                  cmp_func=line_cmp_vert,
                                                  eq_func=line_eq_vert)

        if closest_node is not None:
            logging.debug("Closest Node: {}".format(closest_node.value.index))
        else:
            logging.debug("Closest Node is None")
        return closest_node, d



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging


class VoronoiGraph:

    @staticmethod
    def build(data=None, rerun=False, random=None):
        """ Create a graph of initial random sites """
        logging.debug("Initialising graph")
        self.reset()

        if random is None:
            random = base_random.random

        values = data
        if values is None and not rerun:
            values = self.load_graph()

        assert(values is None or isinstance(values, np.ndarray))
        #create a (n, 2) array of coordinates for the sites, if no data has been loaded
        if values is None:
            logging.debug("Generating values")
            for _ in range(self.node_size):
                rnd_amnt = random((1, 2))
                #scale the new site
                scaler = self.bbox.reshape((2, 2)).transpose()
                new_site = scaler[:, 0] + (rnd_amnt * (scaler[:, 1] - scaler[:, 0]))
                if values is None:
                    values = new_site
                else:
                    values = np.row_stack((values, new_site))

        #setup the initial site events:
        used_coords = []
        for site in values:
            #Avoid duplications:
            if (site[0], site[1]) in used_coords:
                logging.warning("Skipping Duplicate: {}".format(site))
                continue
            #Create an empty face for the site
            future_face = self.dcel.newFace(site, data=BASE_VORONOI_FACE_DATA)
            event = SiteEvent(site, face=future_face)
            heapq.heappush(self.events, event)
            self.sites.append(event)
            used_coords.append((site[0], site[1]))

        #Save the nodes
        if not rerun:
            self.save_graph(values)
        return values



#!/usr/bin/env python3
"""

"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1

##-- end builtin imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from tomlguard.utils.proxy import TomlGuardProxy
from tomlguard.utils import iter_proxy
from tomlguard.error import TomlAccessError

class GuardProxyEntryMixin:
    """ A Mixin to add to GuardBase.
    enables handling a number of conditions when accessing values in the underlying data.
    eg:
    tg.on_fail(2, int).a.value() # either get a.value, or 2. whichever returns has to be an int.
    """

    def on_fail(self, fallback:Any, types:Any|None=None) -> TomlGuardProxy:
        """
        use a fallback value in an access chain,
        eg: doot.config.on_fail("blah").this.doesnt.exist() -> "blah"

        *without* throwing a TomlAccessError
        """
        index = self._index()
        if index != ["<root>"]:
            raise TomlAccessError("On Fail not declared at entry", index)

        return TomlGuardProxy(self, types=types, fallback=fallback)

    def first_of(self, fallback:Any, types:Any|None=None) -> TomlGuardIterProxy:
        """
        get the first non-None value from a index path, even across arrays of tables
        so instead of: data.a.b.c[0].d
        just:          data.first_of().a.b.c.d()
        """
        index = self._index()[:]

        if index != ["<root>"]:
            raise TomlAccessError("Any Of not declared at entry", index)

        return iter_proxy.TomlGuardIterFirstProxy(self, fallback=fallback, types=types)

    def all_of(self, fallback:Any, types:Any|None=None) -> TomlGuardIterProxy:
        raise NotImplementedError()
        index = self._index()[:]

        if index != ["<root>"]:
            raise TomlAccessError("All Of not declared at entry", index)

        return iter_proxy.TomlGuardIterAllProxy(self, fallback=fallback)

    def flatten_on(self, fallback:Any) -> TomlGuardIterProxy:
        """
        combine all dicts at the call site to form a single dict
        """
        raise NotImplementedError()
        if not isinstance(fallback, (type(None), dict)):
            raise TypeError()

        index = self._index()

        if index != ["<root>"]:
            raise TomlAccessError("Flatten On not declared at entry", index)

        return iter_proxy.TomlGuardIterFlatProxy(self, fallback=fallback)

    def match_on(self, **kwargs:tuple[str,Any]) -> TomlGuardIterProxy:
        raise NotImplementedError()
        index = self._table()[:]
        if index != ["<root>"]:
            raise TomlAccessError("Match On not declared at entry", index)
        return iter_proxy.TomlGuardIterMatchProxy(self, fallback=kwargs)



"""
	Constant strings for data held in face/halfedge/vertex .data fields
"""
from typing import Final
from enum import Enum

FaceE              : Final = Enum("Face Data Enum", "FILL TEXT TEXT_OFFSET STROKE CENTROID CEN_RADIUS NULL STARTVERT STARTRAD WIDTH SAMPLE SAMPLE_POST")
EdgeE              : Final = Enum("HalfEdge Data Enum", "STROKE WIDTH START END STARTRAD ENDRAD NULL TEXT BEZIER SAMPLE BEZIER_SIMPLIFY SAMPLE_POST")
VertE              : Final = Enum("Vertex Data Enum", "STROKE RADIUS NULL TEXT SAMPLE SAMPLE_POST")

EditE              : Final = Enum("Edit return type enum", "MODIFIED NEW")
SampleE            : Final = Enum("Ways to sample a drawing", "CIRCLE VECTOR TARGET ANGLE TRANSFER")
SampleFormE        : Final = Enum("How the SampleFunction should treat the target", "FACE EDGE VERTEX")

EDGE_FOLLOW_GUARD  : Final = 200

#The amount used to nudge forwards/down the sweep line in the y direction
#for line segment intersection in dcel.intersect_halfedges
#Default: -0.1 for cartesian bboxs of larger than 0-1
SWEEP_NUDGE        : Final = - 1e-3




#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class BeachLine:

    def get_chain(self):
        """ Get the sequence of leaf values, from left to right """
        if self.root is None:
            return []
        found = set()
        chain = []
        current = self.root.min()
        while current is not None:
            if current in found:
                raise Exception("This shouldn't happen")
            found.add(current)
            chain.append(current)
            current = current.get_successor()
        return chain

    def get_successor_triple(self, node):
        """ Get the trio of this node and two successors """
        if node is None:
            return None
        a = node
        b = a.get_successor()
        if b is not None:
            c = b.get_successor()
            if c is not None:
                return (a, b, c)
        return None

    def get_predecessor_triple(self, node):
        """ Get the trio of this node and two predecessors """
        if node is None:
            return None
        a = node
        b = a.getPredecessor()
        if b is not None:
            c = b.getPredecessor()
            if c is not None:
                return (c, b, a)
        return None

    def insert_successor(self, existing_node, new_value, data=None):
        """ Insert a new value as a successor to a specific node """
        assert(existing_node is None or isinstance(existing_node, Node))
        logging.debug("Inserting Successor")
        new_node = Node(new_value, data=data)
        self.nodes.append(new_node)
        if existing_node is None:
            self.root = new_node
        else:
            existing_node.add_right(new_node)
        self.__balance(new_node)
        return new_node

    def insert_predecessor(self, existing_node, new_value, data=None):
        """ Insert a new value as a predecessor to the specific node """
        assert(existing_node is None or isinstance(existing_node, Node))
        logging.debug("Inserting predecessor")
        new_node = Node(new_value, data=data)
        self.nodes.append(new_node)
        if existing_node is None:
            self.root = new_node
        else:
            existing_node.add_left(new_node)
        self.__balance(new_node)
        return new_node



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from functools import partial
from math import atan2, copysign

import cairo_utils as cu
import numpy as np
from cairo_utils.constants import (BEZIER_DATA_LEN, CIRCLE_DATA_LEN, DELTA,
                                   EPSILON, HALFDELTA, LINE_DATA_LEN,
                                   NODE_RECIPROCAL, PI, QUARTERPI,
                                   SAMPLE_DATA_LEN, TOLERANCE, TWOPI,
                                   IntersectEnum)
from numpy import cos, pi, sin
from scipy.interpolate import splev, splprep

def bbox_to_lines(bbox, epsilon=EPSILON):
    """ take in the min and max values of a bbox,
    return back a list of 4 lines with the enum designating their position """
    assert(isinstance(bbox, np.ndarray))
    assert(bbox.shape == (4, ))
    bbox_e = bbox + np.array([-epsilon, -epsilon, epsilon, epsilon])
    # [[minx, miny], [maxx, maxy]] -> [[minx, maxx], [miny, maxy]]
    bbox_t = bbox_e.reshape((2, 2)).transpose()
    #convert the bbox to bounding lines
    select_x = np.array([1, 0])
    select_y = np.array([0, 1])
    mins = bbox_t[:, 0]
    maxs = bbox_t[:, 1]
    min_x_max_y = mins * select_x + maxs * select_y
    max_x_min_y = maxs * select_x + mins * select_y
    lines = [(np.row_stack((mins, max_x_min_y)), IntersectEnum.HBOTTOM),
             (np.row_stack((min_x_max_y, maxs)), IntersectEnum.HTOP),
             (np.row_stack((mins, min_x_max_y)), IntersectEnum.VLEFT),
             (np.row_stack((max_x_min_y, maxs)), IntersectEnum.VRIGHT)]

    return lines

def bound_line_in_bbox(line, bbox):
    """ takes in a line, limits it to be within a bbox """
    #replace original line endpoint with intersection point
    bbl = bbox_to_lines(bbox)
    intersections = [x for x in [intersect(line, x) for x, y in bbl] if x is not None]
    if not bool(intersections):
        return [line]
    return [np.array([line[0], x]) for x in intersections]

def calc_bbox_corner(bbox, ies, epsilon=EPSILON):
    """ Calculate the nearest corner of a bbox for set of existing intersections  """
    assert(isinstance(bbox, np.ndarray))
    assert(bbox.shape == (4, ))
    assert(isinstance(ies, set))
    hb = IntersectEnum.HBOTTOM
    ht = IntersectEnum.HTOP
    vl = IntersectEnum.VLEFT
    vr = IntersectEnum.VRIGHT
    bbox_e = bbox + np.array([-epsilon, -epsilon, epsilon, epsilon])
    # [[minx, miny], [maxx, maxy]] -> [[minx, maxx], [miny, maxy]]
    bbox_t = bbox_e.reshape((2, 2)).transpose()
    #convert the bbox to bounding lines
    select_x = np.array([1, 0])
    select_y = np.array([0, 1])
    mins = bbox_t[:, 0]
    maxs = bbox_t[:, 1]
    min_x_max_y = mins * select_x + maxs * select_y
    max_x_min_y = maxs * select_x + mins * select_y

    if ies.issubset([hb, vl]):
        return mins
    elif ies.issubset([hb, vr]):
        return max_x_min_y
    elif ies.issubset([ht, vl]):
        return min_x_max_y
    elif ies.issubset([ht, vr]):
        return maxs
    else:
        raise Exception("Calculating box corner failed for: {}".format(ies))

def bbox_centre(bbox):
    """ Get the centre of a bbox """
    assert(isinstance(bbox, np.ndarray))
    assert(bbox.shape == (4, ))
    bbox_t = bbox.reshape((2, 2)).transpose()
    mins = bbox_t[:, 0]
    maxs = bbox_t[:, 1]
    ranges = maxs - mins
    mid = ranges * 0.5
    return mid

def make_bbox_from_point(point, i):
    """ Given a centre point and a length, create a bbox """
    border = HALFDELTA * (1/(i+1))
    return [point[0] - border, point[1] - border, point[0] + border, point[1] + border]

def within_bbox(point, bbox, tolerance=TOLERANCE):
    """ Test whether a point is within the given bbox """
    assert(isinstance(bbox, np.ndarray))
    assert(bbox.shape == (4, ))
    assert(isinstance(point, np.ndarray))
    assert(point.shape == (2, ))
    mod_bbox = bbox + np.array([-tolerance, -tolerance, tolerance, tolerance])
    in_x_bounds = mod_bbox[0] < point[0] and point[0] < mod_bbox[2]
    in_y_bounds = mod_bbox[1] < point[1] and point[1] < mod_bbox[3]
    return in_x_bounds and in_y_bounds



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import mmap

class JGDVMemMap:
    """
      A utility wrapper to easily make memory mapped files from a path,
      and use them as in a context manager
    """

    def __init__(self, path:pl.Path):
        self._path = path
        self._file = None
        self._mmap = None

    def __enter__(self):
        if self._mmap:
            return

        self._file = open(self._path, "r+b")
        self._mmap = mmap.mmap(self._file.fileno(), 0)

    def __exit__(self, exc_type, exc_value, exc_traceback) -> bool:
        if self._mmap:
            self._mmap.close()
        if self._file:
            self._file.close()

        return

    def seek(self, x):
        raise NotImplementedError()

    def read(self):
        raise NotImplementedError()



#/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import abc
import logging as logmod
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from re import Pattern
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

import pyparsing as pp
import pyparsing.core as ppc

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

MARK                      = ">!<"
MATCHING                  = "Matching" + (" " * 17)
MATCHED                   = "Matched"  + ("-" * 14)
FAILED                    = "Failed"   + ("_" * 11)

try_format   : Final[str] = "{3[0]:>10}{3[1]:3}{3[2]:<10} {0} <{1}> at {2}:  "
match_format : Final[str] = "{4[0]:>10}{4[1]:3}{4[2]:<10}\t {0} <{1}> ({2}) {3}"
fail_format  : Final[str] = "{5[0]:>10}{5[1]:3}{5[2]:<10}\t\t {0} <{1}>: {2} found '{3}' at {4}"

class PyParsingDebuggerControl:

    @staticmethod
    def debug_pyparsing(*flags, all_warnings=False):
        """ Set pyparsing to debug
        Only applies for parsers created *after* this,
        so has to be set at boot time.
        """
        # pylint: disable=protected-access
        if flags is None:
            flags = [pp.Diagnostics.enable_debug_on_named_expressions]
        if not debug_pyparsing_active_p():
            logging.info("Enabling Debug on %s Parsers", "Named")
            pp.__diag__.enable_debug_on_named_expressions = True
            if bool(flags):
                for flag in flags:
                    pp.enable_diag(flag)

            if all_warnings:
                pp.enable_all_warnings()
            ppc._default_success_debug_action                 = debug_match_action
            ppc._default_start_debug_action                   = debug_try_action
            ppc._default_exception_debug_action               = debug_fail_action
        else:
            logging.warning("PyParsing Debug is already active")


    @staticmethod
    def dfs_activate(*parsers, remove=False):
        """ DFS on a parser, adding debug funcs to named sub parsers """
        # pylint: disable=protected-access
        queue = list(parsers)
        found = set()
        while bool(queue):
            current = queue.pop(0)
            if current is None or id(current) in found:
                continue

            found.add(id(current))

            if bool(current.name) and current.name != current._defaultName and not remove:
                current.set_debug_actions(debug_try_action,
                                        debug_match_action,
                                        debug_fail_action)
            elif remove:
                current.set_debug_actions(None, None, None)

            if hasattr(current, 'expr'):
                queue.append(current.expr)
            elif hasattr(current, 'exprs'):
                queue += current.exprs

    @static
    def debug_active_p() -> bool:
        return pp.__diag__.enable_debug_on_named_expressions


def debug_try_action(instring, loc, expr, *args):
    """
    Log Entry into parsers
    """
    # pylint: disable=unused-argument, consider-using-f-string
    context = _calc_mark_string(instring, loc)
    logging.warning(try_format.format(MATCHING, expr.name, loc, context))

def debug_match_action(instring, startloc, endloc, expr, toks, *args):
    """
    Log Parser Success
    """
    # pylint: disable=unused-argument, consider-using-f-string
    context = _calc_mark_string(instring, endloc)
    logging.warning(match_format.format(MATCHED, expr.name, str(toks), endloc, context))

def debug_fail_action(instring, loc, expr, exc, *args):
    """
    Log Parser failure
    """
    # pylint: disable=unused-argument, consider-using-f-string
    if isinstance(exc, pp.ParseBaseException):
        found_str = exc.pstr[exc.loc:exc.loc + 1].replace(r'\\', '\\').replace("\n", "\\n")
        mark_str  = _calc_mark_string(instring, exc.loc)
        msg       = exc.msg
        loc       = exc.loc
    else:
        found_str = "AssertionError"
        mark_str  = ("", "", "")
        msg       = ""
        loc       = ""

    logging.error(fail_format.format(FAILED, expr.name, msg, found_str, loc, mark_str))







## util
def _calc_mark_string(instring, loc, buffer=10):
    str_len  = len(instring)
    pre_str  = instring[max(0, loc-buffer):max(0, loc)]
    post_str = instring[max(0, loc):min(str_len, loc+buffer)]
    return pre_str.replace("\n", "\\n"), MARK, post_str.replace("\n", "\\n")



# except pp.ParseException as err:
#     import traceback
#     traceback.print_tb(err.__traceback__)
#     logging.warning(f"Parse Failure: {err.markInputline()}")



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from jgdv._interfaces.decorator import JGDVDecorator_i

DEBUG_DESTRUCT_ON = False

class LogDestruction(JGDVDecorator_i):
    """
    A Decorator to log when instances of a class are deleted
    """

    def _debug_del(self):
        """ standalone del logging """
        logging.warning("Deleting: %s", self)

    def _debug_del_dec(fn):
        """ wraps existing del method """
        def _wrapped(*args):
            logging.warning("Deleting: %s", self)
            fn(*args)

    def __call__(self):
        """
        A Class Decorator, attaches a debugging statement to the object destructor
        """
        match (DEBUG_DESTRUCT_ON, hasattr(cls, "__del__")):
            case (False, _):
                pass
            case (True, True):
                setattr(cls, "__del__", self._debug_del_dec(cls.__del__))
            case (True, False):
                setattr(cls, "__del__", self._debug_instance_del)
        return cls



#!/usr/bin/env python3
"""
Alternative doot cli runner, to use the doot loader
"""
##-- imports
from __future__ import annotations

import abc
import logging as logmod
import pathlib
import pathlib as pl
import sys
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from re import Pattern
from sys import stderr, stdout
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref
##-- end imports

logging         = logmod.root
printer         = logmod.getLogger("doot._printer")

import sh
import stackprinter
import tomlguard as TG
from bdb import BdbQuit

def main():
    result  = 1
    errored = False
    overlord = None
    try:
        log_config = DootLogConfig()
        # --- Setup
        if not bool(doot.config):
            doot.setup()

        log_config.setup()

        logging.info("Called with: %s", sys.argv)
        from jgdv.importing.plugin_loader import JGDVPluginLoader
        from doot.control.overlord import DootOverlord
        overlord  = DootOverlord(loaders={"plugin": JGDVPluginLoader().setup(sys.argv[:]) },
                                 config_filenames=[doot.constants.paths.DEFAULT_LOAD_TARGETS],
                                 log_config=log_config,
                                 args=sys.argv[:])

        # --- Do whatever thats been triggered
        result    = overlord()

    ##-- handle doot errors
    except (doot.errors.DootEarlyExit, BdbQuit):
        printer.warning("Early Exit Triggered")
    except doot.errors.DootMissingConfigError as err:
        # Handle missing files
        if not doot.constants.DEFAULT_LOAD_TARGETS[0].exists():
            if input("No toml config data found, create stub doot.toml? _/n ") != "n":
                doot.constants.DEFAULT_LOAD_TARGETS[0].write_text(doot.constants.TOML_TEMPLATE.read_text())
                logging.info("Stubbed")

    except doot.errors.DootTaskError as err:
        errored = True
        printer.error("%s : %s", err.general_msg, err.task_name)
        printer.error("---- Source: %s", err.task_source)
        printer.error("---- %s", str(err))
    except doot.errors.DootError as err:
        errored = True
        printer.error("%s", err.general_msg)
        printer.error("---- %s", str(err))
    ##-- end handle doot errors
    ##-- handle todo errors
    except NotImplementedError as err:
        errored = True
        logging.error(stackprinter.format())
        printer.error("Not Implemented: %s", err)
    ##-- end handle todo errors
    ##-- handle general errors
    except Exception as err:
        errored = True
        logging.error(stackprinter.format())
        # logging.error("Python Error: %s", err)
    ##-- end handle general errors
    finally: # --- final shutdown
        if overlord:
            overlord.shutdown()
        match sys.platform:
            case "linux":
                pass
            case "darwin":
                announce_exit : bool = doot.constants.ANNOUNCE_EXIT
                announce_voice : str = doot.constants.ANNOUNCE_VOICE
                if doot.config is not None:
                    announce_exit        = doot.config.on_fail(announce_exit, bool|str).settings.general.notify.say_on_exit()
                    announce_voice       = doot.config.on_fail(announce_voice, str).setttings.general.notify.announce_voice()

                match errored, announce_exit:
                    case False, str() as say_text:
                        cmd = sh.say("-v", announce_voice, "-r", "50", say_text)
                    case False, True:
                        cmd = sh.say("-v", announce_voice, "-r", "50", "Doot Has Finished")
                    case True, True|str():
                        cmd = sh.say("-v", announce_voice, "-r", "50", "Doot Encountered a problem")
                    case _:
                        cmd = None
                if cmd is not None:
                    cmd.execute()

        sys.exit(result)





#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from functools import partial
from math import atan2, copysign

import cairo_utils as cu
import numpy as np
from cairo_utils.constants import (BEZIER_DATA_LEN, CIRCLE_DATA_LEN, DELTA,
                                   EPSILON, HALFDELTA, LINE_DATA_LEN,
                                   NODE_RECIPROCAL, PI, QUARTERPI,
                                   SAMPLE_DATA_LEN, TOLERANCE, TWOPI,
                                   IntersectEnum)
from numpy import cos, pi, sin
from scipy.interpolate import splev, splprep


def sort_coords(arr):
    """ Sort a list of points by x then y value  """
    ind = np.lexsort((arr[:, 1], arr[:, 0]))
    return arr[ind]



#!/usr/bin/env python3
"""
This is separated from the rest of the handler_system to avoid a circular import

"""
# pylint: disable=abstract-method,invalid-sequence-index,use-a-generator,too-many-lines
# pyright: reportPrivateUsage=warning
##-- imports
from __future__ import annotations

import logging as logmod
from dataclasses import InitVar, dataclass, field, replace
from enum import Enum
from types import MethodType
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Generic, Iterable,
                    Iterator, Mapping, Match, MutableMapping, NewType,
                    Protocol, Sequence, Tuple, Type, TypeAlias, TypeVar, cast)

##-- end imports

logging     = logmod.getLogger(__name__)

@APE.assert_concrete
@dataclass
class BasicHandler(HS.Handler_i):

    def __post_init__(self) -> None:
        if isinstance(self.func, type):
            self.func = self.func()

        if self.func is not None:
            assert(callable(self.func))

        if self.struct is not None or self.struct_i is None:
            pass
        elif hasattr(self.struct_i, "build_default"):
            self.struct = self.struct_i.build_default() #type:ignore
        else:
            self.struct = self.struct_i() #type:ignore

    def __call__(self, *args, **kwargs):
        if self.func is None:
            raise AcabHandlerException("Attempt to Call Struct Handler", rest=[self])
        return self.func(*args, **kwargs)

    def __iter__(self):
        """ unpack the handler"""
        return (self.func, self.struct).__iter__()

    def __repr__(self):
        sig_s       = str(self.signal)
        func_name   = ""
        struct_name = ""
        if self.func is not None:
            func_name = str(self.func.__class__.__name__) #type:ignore
        if self.struct is not None:
            struct_name = str(self.struct.__class__.__name__)

        return f"<{self.__class__.__name__}({sig_s}: {func_name}: {struct_name})>"

    def verify(self, instruction):
        result = False
        if self.verify_f is not None:
            result = self.verify_f(instruction)
        else:
            result = True

        if not result and self.func is not None and hasattr(self.func, "verify"):
            result = result and self.func.verify(instruction) #type:ignore

        return result

    def as_handler(self, *, signal=None, struct=None, flags=None):
        """
        Duplicate the handler, with some changes
        """
        return self.__class__(signal or self.signal,
                              func=self.func,
                              struct=struct or self.struct,
                              flags=flags or self.flags)

    @cache
    def __str__(self):
        return str(self.signal)



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@dataclass
class IntersectorState:
    """ Processes a DCEL to intersect half_edges,
    in a self contained class
    """

    dcel            : 'DCEL'         = field()
    edge_set        : Set[HalfEdge]  = field(init=False, default_factory=set)
    lower_edges     : List[HalfEdge] = field(init=False, default_factory=list)
    results         : List[Any]      = field(init=False, default_factory=list)
    discovered      : Set[Any]       = field(init=False, default_factory=set)
    sweep_y         : float          = field(init=False, default=inf)
    sweep_y_prev    : float          = field(init=False, default=inf)
    #Tree to keep active edges in,
    #Sorted by the x's for the current y of the sweep line
    status_tree     : RBTree         = field(init=False, default=None)
    #Heap of (vert, edge) pairs,
    #with invariant : all([e.is_upper() for v, e in event_list])
    event_list      : List[Event]    = field(init=False, default_factory=list)

    def __init__(self, dcel):
        self.status_tree = RBTree(cmp_func=line_cmp, eq_func=line_eq)



#!/usr/bin/env python3
"""
Pytest testing fixtures
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import pytest
import os

##-- imports
import tempfile
##-- end imports

# with tempfile.TemporaryDirectory() as temp_dir:
#      assert(isisntance(temp_dir, str))
#      # Do Stuff

# On Exit of context manager, dir is deleted

def _add_prop(m, name, val):
    setattr(type(m), name, PropertyMock(return_value=val))


def filter_str_warns():
    warnings.warn(DeprecationWarning("invalid escape sequence"))
    return 1

# @pytest.mark.filterwarnings("ignore:filter_str_warns")
# def test_...


@pytest.fixture
def wrap_tmp(tmp_path):
    """ create a new temp directory, and change cwd to it,
      returning to original cwd after the test
      """
    logging.debug("Moving to temp dir")
    orig     = pl.Path().cwd()
    new_base = tmp_path / "test_root"
    new_base.mkdir()
    os.chdir(new_base)
    doot.locs._root = new_base
    yield new_base
    logging.debug("Returning to original dir")
    os.chdir(orig)
    doot.locs._root = orig



#!/usr/bin/env python3
"""

"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1

##-- end builtin imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import os
from sys import stdout, stderr
from jgdv.logging.log_colour import JGDVColourFormatter, JGDVColourStripFormatter

env : dict = os.environ

class _JGDVAnyFilter:
    """

    """

    def __init__(self, names=None, reject=None):
        self.names      = names or []
        self.rejections = reject or []
        self.name_re    = re.compile("^({})".format("|".join(self.names)))

    def __call__(self, record):
        return (record.name not in self.rejections) and (record.name == "root"
                                                         or not bool(self.names)
                                                    or self.name_re.match(record.name))

class JGDVLogConfig:
    """ Utility class to setup [stdout, stderr, file] logging. """

    def __init__(self):
        # Root Logger for everything
        self.root    = logmod.root
        # EXCEPT this, which replaces 'print(x)'
        self.printer               = logmod.getLogger(doot.constants.printer.PRINTER_NAME)

        self.file_handler          = None
        self.stream_handler        = logmod.StreamHandler(stdout)
        self.print_stream_handler  = logmod.StreamHandler(stdout)

        self._pre_setup()

    def _pre_setup(self):
        """ a basic, config-less setup """
        self.root.setLevel(logmod.NOTSET)
        # self.file_handler.setFormatter(JGDVColourStripFormatter(fmt="{levelname} : INIT : {message}"))

        self.stream_handler.setLevel(logmod.WARNING)
        self.stream_handler.setFormatter(logmod.Formatter("{levelname}  : INIT : {message}", style="{"))

        self.root.addHandler(self.stream_handler)

        self.printer.propagate = False
        self.print_stream_handler.setFormatter(logmod.Formatter("{message}", style="{"))
        self.printer.setLevel(logmod.NOTSET)
        self.printer.addHandler(self.print_stream_handler)
        # self.printer.addHandler(self.file_handler)

    def setup(self):
        """ a setup that uses config values """
        assert(doot.config is not None)
        self._setup_file_logging()
        self._setup_stream_logging()
        self._setup_print_logging()

    def _setup_file_logging(self):
        log_name_format = doot.config.on_fail("doot-%Y-%m-%d::%H:%M.log").logging.file.name()
        log_file_name   = datetime.datetime.now().strftime(log_name_format)

        log_dir = doot.locs["{logs}"]
        # TODO delete old logs if number of logs larger than N
        if not log_dir.exists():
            log_dir = pl.Path()

        log_file_path = log_dir / log_file_name

        self.file_handler  = logmod.FileHandler(log_file_path, mode='w')
        file_log_level    = doot.config.on_fail("DEBUG", str|int).logging.file.level(wrapper=lambda x: logmod._nameToLevel.get(x, 0))
        file_log_format   = doot.config.on_fail("{levelname} : {pathname} : {lineno} : {funcName} : {message}", str).logging.file.format()
        file_filter_names = doot.config.on_fail([], list).logging.file.allow()

        self.file_handler.setLevel(file_log_level)
        self.file_handler.setFormatter(JGDVColourStripFormatter(fmt=file_log_format))
        if bool(file_filter_names):
            self.file_handler.addFilter(_JGDVAnyFilter(file_filter_names))

        self.root.addHandler(self.file_handler)

    def _setup_stream_logging(self):
        stream_log_level    = doot.config.on_fail("WARNING", str|int).logging.stream.level(wrapper=lambda x: logmod._nameToLevel.get(x, 0))
        stream_log_format   = doot.config.on_fail("{levelname} : {pathname} : {lineno} : {funcName} : {message}", str).logging.stream.format()
        stream_filter_names = doot.config.on_fail([], list).logging.stream.allow()

        self.stream_handler.setLevel(stream_log_level)
        use_colour = doot.config.on_fail(False, bool).logging.stream.colour()
        use_colour &= "PRE_COMMIT" not in env
        if use_colour:
            self.stream_handler.setFormatter(JGDVColourFormatter(fmt=stream_log_format))
        else:
            self.stream_handler.setFormatter(JGDVColourStripFormatter(fmt=stream_log_format))

        if bool(stream_filter_names):
            self.stream_handler.addFilter(_JGDVAnyFilter(stream_filter_names))

    def _setup_print_logging(self):
        printer_log_level    = doot.config.on_fail("NOTSET", str|int).logging.printer.level(wrapper=lambda x: logmod._nameToLevel.get(x, 0))
        printer_log_format   = doot.config.on_fail("{message}", str).logging.printer.format()

        self.print_stream_handler.setLevel(printer_log_level)

        use_colour = doot.config.on_fail(False, bool).logging.printer.colour()
        use_colour &= "PRE_COMMIT" not in env
        if use_colour:
            self.print_stream_handler.setFormatter(JGDVColourFormatter(fmt=printer_log_format))
        else:
            self.print_stream_handler.setFormatter(JGDVColourStripFormatter(fmt=printer_log_format))

    def set_level(self, level):
        self.stream_handler.setLevel(level)



#!/usr/bin/env python3
"""
A Proxy for TomlGuard,
  which allows you to use the default attribute access
  (data.a.b.c)
  even when there might not be an `a.b.c` path in the data.

  Thus:
  data.on_fail(default_value).a.b.c()

  Note: To distinguish between not giving a default value,
  and giving a default value of `None`,
  wrap the default value in a tuple: (None,)
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeVar, NoReturn,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1

##-- end builtin imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from types import UnionType
from tomlguard.utils.trace_helper import TraceHelper
from tomlguard.base import GuardBase
from tomlguard.error import TomlAccessError
from tomlguard.base import TomlTypes

NullFallback = NoReturn

class TomlGuardProxy:
    """
    A Wrapper for guarded access to toml values.
    you get the value by calling it.
    Until then, it tracks attribute access,
    and reports that to GuardBase when called.
    It also can type check its value and the value retrieved from the toml data
    """

    def __init__(self, data:GuardBase, types:Any=None, index:list[str]|None=None, fallback:TomlTypes|NullFallback=NullFallback):
        self._types                         = types or Any
        self._data                          = data
        self.__index : list[str]            = index or ["<root>"]
        if fallback == (None,):
            self._fallback = None
        else:
            self._fallback = fallback

        if fallback:
            self._match_type(self._fallback)

    def __repr__(self) -> str:
        type_str = self._types_str()
        index_str = ".".join(self._index())
        return f"<TomlGuardProxy: {index_str}:{type_str}>"

    def __call__(self, wrapper:callable[[TomlTypes], Any]|None=None, fallback_wrapper:callable[[TomlTypes], Any]|None=None) -> Any:
        """
        Reify a proxy into an actual value, or its fallback.
        Optionally call a wrapper function on the actual value,
        or a fallback_wrapper function on the fallback
        """
        self._notify()
        wrapper : callable[[TomlTypes], TomlTypes] = wrapper or (lambda x: x)
        fallback_wrapper                           = fallback_wrapper or (lambda x: x)
        match self._data, self._fallback:
            case x, y if x is NullFallback and y is NullFallback:
                raise ValueError("No Value, and no fallback")
            case x, None if x is NullFallback or x is None:
                val = None
            case x, data if x is NullFallback or x is None:
                val = fallback_wrapper(data)
            case GuardBase() as data, _:
                val = wrapper(dict(data))
            case _ as data, _:
                val = wrapper(data)

        return self._match_type(val)

    def __getattr__(self, attr:str) -> TomlGuardProxy:
        try:
            match self._data:
                case x if x is NullFallback:
                    raise TomlAccessError()
                case GuardBase():
                    return self._inject(self._data[attr], attr=attr)
                case _:
                    return self._inject(attr=attr)
        except TomlAccessError:
            return self._inject(clear=True, attr=attr)

    def _inject(self, val:tuple[Any]=NullFallback, attr:str|None=None, clear:bool=False) -> TomlGuardProxy:
        match val:
            case _ if clear:
                val = NullFallback
            case x if x is NullFallback:
                val = self._data
            case _:
                pass

        return TomlGuardProxy(val,
                              types=self._types,
                              index=self._index(attr),
                              fallback=self._fallback)


    def __getitem__(self, keys:str|tuple[str]) -> TomlGuardProxy:
        curr = self
        match keys:
            case tuple():
                for key in keys:
                    curr = curr.__getattr__(key)
            case str():
                curr = self.__getattr__(keys)

        return curr

    def __len__(self) -> int:
        if hasattr(self._data, "__len__"):
            return len(self._data)

        return 0

    def __bool__(self) -> bool:
        return self._data is not None and self._data is not NullFallback

    def _notify(self) -> None:
        types_str = self._types_str()
        match self._data, self._fallback, self._index():
            case GuardBase(), _, _:
                pass
            case _, _, []:
                pass
            case x , val, [*index] if x is NullFallback:
                GuardBase.add_defaulted(".".join(index), val, types_str)
            case val, _, [*index]:
                GuardBase.add_defaulted(".".join(index), val, types_str)
            case val, flbck, index,:
                raise TypeError("Unexpected Values found: ", val, index, flbck)

    def _types_str(self) -> str:
        match self._types:
            case UnionType() as targ:
                types_str = repr(targ)
            case type(__name__=targ):
                types_str = targ
            case _ as targ:
                types_str = str(targ)

        return types_str

    def _match_type(self, val:TomlTypes) -> TomlTypes:
        if self._types != Any and not isinstance(val, self._types):
            types_str = self._types_str()
            index_str  = ".".join(self.__index + ['(' + types_str + ')'])
            err = TypeError("TomlProxy Value doesn't match declared Type: ", index_str, val, self._types)
            raise err.with_traceback(TraceHelper()[5:10])

        return val

    def _index(self, sub:str=None) -> list[str]:
        if sub is None:
            return self.__index[:]
        return self.__index[:] + [sub]



#!/usr/bin/env python3
# pylint: disable=no-member
##-- imports
from __future__ import annotations

import abc
import logging as logmod
from collections import defaultdict
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from os.path import splitext

##-- end imports

import pyparsing as pp

logging = logmod.getLogger(__name__)
SC      = SimpleLogColour

config = acab.config
import_dict = config.imports.specific

@register_class("log")
class ReportCmd:
    """
    Control the Repl log level, format,
    and specify filters for the logger

    """

    def __init__(self):
        self._parser = self._gen_parser()

    def _gen_parser(self):
        operator_kw  = pp.MatchFirst([pp.Keyword("ops"),
                                      pp.Keyword("operator")])("operator")
        module_kw    = pp.MatchFirst([pp.Keyword("mod"),
                                      pp.Keyword("module")])("module")
        semantic_kw  = pp.Keyword("semantics")("semantics")
        printer_kw   = pp.Keyword("printers")("printers")
        dsl_kw       = pp.Keyword("dsl")("dsl")
        engine_kw    = pp.Keyword("engine")("engine")

        eol          = pp.line_end("all")

        report_parser = pp.ZeroOrMore(pp.MatchFirst([operator_kw,
                                                    module_kw,
                                                    semantic_kw,
                                                    printer_kw,
                                                    dsl_kw,
                                                    engine_kw,
                                                    eol]))
        return report_parser

    def __call__(self, line):
        self.do_report(line)

    def _log_level(self, line):
        """
        Change the console logging level

        Usage:
        log {DEBUG | INFO | WARN | ERROR}
        log
        """
        try:
            root = logmod.getLogger('')
            handler = [x for x in root.handlers if not isinstance(x, logmod.FileHandler)][0]
            if bool(line):
                level = logmod._nameToLevel[line.upper()]
                if root.level > level:
                    logging.warning("Logging Root level is masking the handler level")
                handler.setLevel(level)
                print(f"Set Console Log level to: {line.upper()} : {level}")
            else:
                level = handler.level
                if handler.level in logmod._levelToName:
                    level = logmod._levelToName[handler.level]
                print(f"Console Log Level: {level}")

        except KeyError as err:
            print(f"Unrecognised Log Level: {line.upper()}")

    def _log_fmt(self, line):
        """
        Change the console Log format.
        The Repl Uses AcabLogFormatter and AcabLogRecord by default,
        which uses {} style format strings for the log format
        """
        root = logmod.getLogger('')
        handler = [x for x in root.handlers if not isinstance(x, logmod.FileHandler)][0]
        if bool(line):
            handler.setFormatter(AcabLogFormatter(fmt=line, record=True))
            print(f"Set Console Log Format to: {line}")
        else:
            fmt = handler.formatter._fmt
            recordFactory = logmod.getLogRecordFactory()
            print(f"Console Log Level: {fmt}")
            if hasattr(recordFactory, "available_fields"):
                print(f"Availble Log Fields: {recordFactory.available_fields}")
            else:
                print("For Available Fields see:",
                    "https://docs.python.org/3/library/logging.html#logrecord-attributes")

    def _log_filter(self, line):
        """ Add or remove a logging filter """
        root = logmod.getLogger('')
        handler = root.handlers[1]
        if bool(line):
            handler.addFilter(logmod.Filter(line))
            print(f"Set Console Log Format to: {line}")
        elif bool(handler.filters):
            handler.removeFilter(handler.filters[-1])
            print(f"Removing Last Console Log Filter")
        else:
            print("No filters set")



#!/usr/bin/env python3
"""

"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

def check_protocol(cls):
    """ Decorator. Check the class implements all its methods / has no abstractmethods """

    abstracts = [x for x in dir(cls) if hasattr(getattr(cls, x), "__isabstractmethod__") and getattr(cls, x).__isabstractmethod__]
    if bool(abstracts):
        raise NotImplementedError(f"Class has Abstract Methods: {cls.__module__} : {cls.__name__} : {abstracts}")

    return cls
